__version__ = "0.8.1"
import os

KERNEL_TYPE = os.environ.get("KERNEL_TYPE", "cuda").lower()
KERAS_BACKEND = os.environ.get("KERAS_BACKEND")
BACKEND = os.environ.get("KERNEL_BACKEND")


if KERAS_BACKEND is not None:
    BACKEND = KERAS_BACKEND.lower()
elif BACKEND is not None:
    os.environ["KERAS_BACKEND"] = BACKEND.lower()
else:
    import keras

    BACKEND = "torch"
    os.environ["KERAS_BACKEND"] = BACKEND
    keras.config.set_backend("torch")
assert KERNEL_TYPE in ["triton", "cuda", "native"]
assert BACKEND in ["torch", "jax", "numpy", "tensorflow"]
from .rwkv7_kernel import get_generalized_delta_rule, get_rnn_generalized_delta_rule
from .rwkv6_kernel import get_rwkv6_kernel
from .mhc_kernel import get_mhc_kernel

generalized_delta_rule, generalized_delta_rule_inference = get_generalized_delta_rule(
    KERNEL_TYPE=KERNEL_TYPE
)
rwkv7_op = generalized_delta_rule
rwkv7_op_inference = generalized_delta_rule_inference

rnn_generalized_delta_rule = get_rnn_generalized_delta_rule(KERNEL_TYPE=KERNEL_TYPE)
rwkv7_op_rnn = rnn_generalized_delta_rule


RWKV6_OP = get_rwkv6_kernel(KERNEL_TYPE=KERNEL_TYPE)
mhc_pre_op, mhc_post_op = get_mhc_kernel(KERNEL_TYPE=KERNEL_TYPE)



import os

# 设置环境变量
os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["KERAS_BACKEND"] = "jax"
os.environ["KERNEL_TYPE"] = "triton"
import time
import jax
import jax.numpy as jnp
import numpy as np
from functools import partial
from rwkv_ops.mhc_kernel.jax_triton_op.mhc_post_op import mhc_post_op as triton_mhc_op
from rwkv_ops.mhc_kernel.native_op import mhc_post_op as native_mhc_op

# ------------------------------------------------------------------
# 1. 配置参数与构造输入
# ------------------------------------------------------------------
B, T, n, C = 128, 1024, 4, 512
dtype = jnp.bfloat16

# JAX 使用 PRNGKey 管理随机性
key = jax.random.PRNGKey(42)
k1, k2, k3, k4 = jax.random.split(key, 4)

# 创建随机输入
layer_out_raw = jax.random.normal(k1, (B, T, C)).astype(jnp.bfloat16)
x_expanded_raw = jax.random.normal(k2, (B, T, n, C)).astype(jnp.bfloat16)
h_post_raw_raw = jax.random.normal(k3, (B, T, n)).astype(jnp.float32)
H_res_raw = jax.random.normal(k4, (B, T, n, n)).astype(jnp.float32)


# ------------------------------------------------------------------
# 2. 工具函数：数值比较
# ------------------------------------------------------------------
def test_is_close(name, x1, x2, atol=1e-2, rtol=1e-2):
    x1_np = np.array(x1, dtype=np.float32)
    x2_np = np.array(x2, dtype=np.float32)
    error = np.abs(x1_np - x2_np)
    print(
        f"平均绝对误差是{error.mean()},平均相对误差是{error.mean() / np.abs(x1_np).mean()}"
    )
    equal_rate = np.sum(error < 1e-8) / np.cumprod(x1_np.shape)[-1] * 100
    print(f"{equal_rate:.2f}%的数据完全一样")
    if np.isnan(x1_np).any() or np.isnan(x2_np).any():
        print(f"❌❌ {name} 存在 NaN ❌❌")
        return

    try:
        np.testing.assert_allclose(x1_np, x2_np, atol=atol, rtol=rtol)
        print(f"✅ {name} 数值一致 (max_diff: {np.abs(x1_np - x2_np).max():.6f})")
    except AssertionError as e:
        print(f"❌ {name} 数值不一致")
        print(e)


# ------------------------------------------------------------------
# 3. 精度校验 (Forward & Backward)
# ------------------------------------------------------------------
print("--- 开始前向测试 ---")
output_triton = triton_mhc_op(layer_out_raw, x_expanded_raw, h_post_raw_raw, H_res_raw)
output_native = native_mhc_op(layer_out_raw, x_expanded_raw, h_post_raw_raw, H_res_raw)


print("\n--- 开始反向测试 ---")


# 定义一个计算 Loss 的函数用于求导
def loss_fn(op_func, l, x, h, H):
    out = op_func(l, x, h, H)
    return jnp.mean(out.astype(jnp.float32) ** 2)


# 使用 jax.value_and_grad 获取 loss 和所有输入的梯度
grad_native_fn = jax.value_and_grad(
    partial(loss_fn, native_mhc_op), argnums=(0, 1, 2, 3)
)
grad_triton_fn = jax.value_and_grad(
    partial(loss_fn, triton_mhc_op), argnums=(0, 1, 2, 3)
)

loss_native, grads_native = grad_native_fn(
    layer_out_raw, x_expanded_raw, h_post_raw_raw, H_res_raw
)
loss_triton, grads_triton = grad_triton_fn(
    layer_out_raw, x_expanded_raw, h_post_raw_raw, H_res_raw
)

test_is_close("前向测试结果:Output", output_native, output_triton)
grad_names = ["layer_out", "x_expanded", "h_post_raw", "H_res"]
for i, name in enumerate(grad_names):
    test_is_close(f"反向测试结果:Gradient: {name}", grads_native[i], grads_triton[i])

# ------------------------------------------------------------------
# 4. 性能基准测试 (Benchmark)
# ------------------------------------------------------------------
print("\n" + "=" * 40)
print("🚀 开始性能基准测试 (Speed Benchmark)")
print("=" * 40)

n_warmup = 10
n_repeat = 100

# 准备编译版本
# 注意：triton_mhc_op 内部已经包含了 triton_call，
# 我们将其放入 jax.jit 以确保整个 Dispatch 过程被优化
jit_native_op = jax.jit(native_mhc_op)
jit_triton_op = jax.jit(triton_mhc_op)


def benchmark_op(op_func, name, l, x, h, H, is_backward=False):
    # 封装函数以便于统一求导或直接执行
    if is_backward:
        # 这里的函数包含前向和反向
        run_fn = jax.jit(jax.grad(partial(loss_fn, op_func), argnums=(0, 1, 2, 3)))
    else:
        run_fn = jax.jit(op_func) if "jit" in name.lower() else op_func

    # 预热 (Warmup)
    for _ in range(n_warmup):
        res = run_fn(l, x, h, H)
        # JAX 是异步执行的，必须 block 确保计算完成
        jax.block_until_ready(res)

    # 计时
    start_time = time.perf_counter()
    for _ in range(n_repeat):
        res = run_fn(l, x, h, H)
        jax.block_until_ready(res)
    end_time = time.perf_counter()

    avg_time_ms = (end_time - start_time) / n_repeat * 1000
    label = "Backward (+Fwd)" if is_backward else "Forward"
    print(f"👉 {name:20s} {label}: {avg_time_ms:.4f} ms")
    return avg_time_ms


# 运行测试
print(f"配置: B={B}, T={T}, n={n}, C={C}, dtype={dtype}")

print("\n--- Forward Speed ---")
# JAX Eager 模式通常非常慢，仅供参考
t_eager_fwd = benchmark_op(
    native_mhc_op, "JAX Eager", layer_out_raw, x_expanded_raw, h_post_raw_raw, H_res_raw
)
t_jit_fwd = benchmark_op(
    jit_native_op,
    "JAX JIT (Native)",
    layer_out_raw,
    x_expanded_raw,
    h_post_raw_raw,
    H_res_raw,
)
t_triton_fwd = benchmark_op(
    jit_triton_op,
    "JAX-Triton JIT",
    layer_out_raw,
    x_expanded_raw,
    h_post_raw_raw,
    H_res_raw,
)

print("\n--- Backward Speed (Forward + Backward) ---")
# 注意：JAX 的反向传播必须在编译函数上进行才有意义
t_jit_bwd = benchmark_op(
    native_mhc_op,
    "JAX JIT (Native)",
    layer_out_raw,
    x_expanded_raw,
    h_post_raw_raw,
    H_res_raw,
    is_backward=True,
)
t_triton_bwd = benchmark_op(
    triton_mhc_op,
    "JAX-Triton JIT",
    layer_out_raw,
    x_expanded_raw,
    h_post_raw_raw,
    H_res_raw,
    is_backward=True,
)

# ------------------------------------------------------------------
# 5. 总结分析
# ------------------------------------------------------------------
print("\n" + "=" * 40)
print("📊 性能提升总结 (Speedup)")
print("=" * 40)


def print_speedup(base_name, base_time, target_name, target_time):
    speedup = base_time / target_time
    print(f"{target_name} vs {base_name}: {speedup:.2f}x faster")


print(">>> Forward:")
print_speedup("JAX JIT Native", t_jit_fwd, "JAX-Triton", t_triton_fwd)

print("\n>>> Backward:")
print_speedup("JAX JIT Native", t_jit_bwd, "JAX-Triton", t_triton_bwd)

print("\n🎉🎉 MHC JAX-Triton 算子校验结束 🎉🎉")



from keras import ops


def fp32_sigmoid(x):
    dtype = x.dtype
    return ops.cast(ops.nn.sigmoid(ops.cast(x, "float32")), dtype)


def stream_distribute(inp, h_post_raw):
    """
    Distribute (1 -> n): 将单流输出分发回多流。
    对齐精度版：强制在 FP32 下进行广播乘法。

    inp: [B, T, C] (BF16)
    H_post: [B, T, n] (FP32)
    """
    # 1. 记录原始类型
    original_dtype = inp.dtype

    # 2. 提升到 FP32 进行运算 (对齐 CUDA 内核内部的 to_float 逻辑)
    # [B, T, 1, C]
    x_fp32 = ops.cast(ops.expand_dims(inp, -2), "float32")

    # [B, T, n, 1]
    w_fp32 = ops.expand_dims(2.0 * fp32_sigmoid(ops.cast(h_post_raw, "float32")), -1)

    # 3. 执行广播乘法
    # 结果为 [B, T, n, C]
    res_fp32 = x_fp32 * w_fp32

    # 4. 转回原始类型 (对齐 CUDA 内核末尾的 to_bf 逻辑)
    return ops.cast(res_fp32, original_dtype)


def stream_mix(inp, M):
    """
    Mix (n -> n): 残差流之间的线性交互。
    inp: [B, T, n, C]
    M: [B, T, n, n] 或 [n, n] (由 sinkhorn_knopp 生成的方阵)
    """
    # 使用 einsum 表达矩阵乘法：M @ inp
    # i,j 是流索引，k 是通道索引
    dtype = inp.dtype
    inp = ops.cast(inp, M.dtype)
    if len(ops.shape(M)) == 2:
        out = ops.einsum("ij,btjk->btik", M, inp)
    else:
        out = ops.einsum("btij,btjk->btik", M, inp)
    return ops.cast(out, dtype)


def mhc_post_op(layer_out, x_expanded, h_post_raw, H_res):
    """
    mHC 后处理融合算子
    输入:
        layer_out: [B, T, C] - 核心层 (Attention/FFN) 处理后的输出
        x_expanded: [B, T, n, C] - 之前的扩展残差流 (Pre-Op 之前的状态)
        H_post: [B, T, n] - 分发权重 (来自 Pre-Op)，2*sigmoid后的数值
        H_res: [B, T, n, n] - 流混合矩阵 (来自 Pre-Op)
    返回:
        x_next: [B, T, n, C] - 更新后的扩展残差流
    """
    # 精度转化
    dtype = x_expanded.dtype
    layer_out = ops.cast(layer_out, "float32")
    x_expanded = ops.cast(x_expanded, "float32")
    h_post_raw = ops.cast(h_post_raw, "float32")
    H_res = ops.cast(H_res, "float32")
    # 1. 在 FP32 下计算混合路径
    x_mixed_f32 = stream_mix(x_expanded, H_res)

    # 2. 在 FP32 下计算增量路径
    x_delta_f32 = stream_distribute(layer_out, h_post_raw)

    # 3. 在 FP32 下完成最后的残差加法
    x_next_f32 = x_mixed_f32 + x_delta_f32

    # 4. 只在最后输出时进行一次 BF16 转换
    # 这一步对应 CUDA 内核中最后的 to_bf()
    return ops.cast(x_next_f32, dtype)


def sinkhorn_knopp(inp, num_iters=20, eps=1e-8):
    """
    将输入矩阵投影为双拟随机矩阵 (Doubly Stochastic Matrix)。
    通常 inp 是 log 域的矩阵 (H_res_raw)。
    """

    # 转换到 fp32 并应用 exp (论文 Eq. 9 之前的步骤)
    x = ops.cast(inp, "float32")
    # 防溢出技巧：减去最大值
    x = x - ops.max(x, axis=(-1, -2), keepdims=True)
    P = ops.exp(x)

    for _ in range(num_iters):
        # 行归一化
        P = P / (ops.sum(P, axis=-1, keepdims=True) + eps)
        # 列归一化
        P = P / (ops.sum(P, axis=-2, keepdims=True) + eps)

    return P


def mhc_rmsnorm(inp, eps=1e-5):
    """
    标准 RMSNorm 算子。
    inp: [..., C], weight: [C]
    """
    dtype = inp.dtype
    x = ops.cast(inp, "float32")
    # 计算均方根
    rms = ops.sqrt(ops.mean(ops.square(x), axis=-1, keepdims=True) + eps)
    x_normed = x / rms
    # 应用权重
    return ops.cast(x_normed, dtype)


def stream_aggregate(inp, H_pre):
    # 1. 转换为 float32 进行高精度计算
    inp_f32 = ops.cast(inp, "float32")
    H_f32 = ops.cast(H_pre, "float32")
    H_f32 = fp32_sigmoid(H_f32)
    # 2. 在 float32 空间完成乘法和累加
    out_f32 = ops.sum(inp_f32 * ops.expand_dims(H_f32, -1), axis=-2)

    # 3. 最后转回原先的格式 (如 bf16)
    return ops.cast(out_f32, inp.dtype)


def linear_and_reshape(
    x_norm,
    alpha_pre,
    alpha_post,
    alpha_res,
    phi,
    bias_pre,
    bias_post,
    bias_res,
    n,
    eps=1e-5,
):
    """
    mHC 动态投影与分支生成算子（底层实现）

    功能：将多流特征 x 通过线性投影生成三组混合系数，用于后续的聚合/分发/残差操作。
    实现：RMSNorm → Linear → 动态拆分为三个分支 → 可选激活

    --- 参数 ---
    x: [batch_size, seq_len, n * hidden_size], BF16/BF16
        展平后的多流特征（n 个并行流，每个维度为 hidden_size）

    alpha_pre, alpha_post, alpha_res: (1,), float32
        三个标量系数，分别控制 H_pre、H_post、H_res 分支的缩放强度

    phi: [n * hidden_size, M], BF16
        投影矩阵，M = n * (n + 2)（包含残差/前/后三个分支的拼接维度）
        **注意**：若 M < 32，CUDA 内核会自动填充到 32 的倍数以保证效率

    bias: [M], float32
        线性层偏置项

    n: int
        扩展率（流数量），用于将输出 reshap 为 n×n 矩阵

    eps: float, default=1e-5
        RMSNorm 的数值稳定常数

    --- 返回 ---
    h_pre_raw: [batch_size, seq_len, n], 原输入 dtype
        原始聚合权重（未激活），需配合 stream_aggregate 使用

    H_post: [batch_size, seq_len, n], 原输入 dtype
        分发权重，已应用 2 * sigmoid 激活，取值范围 (0, 2)

    h_res_reshaped: [batch_size, seq_len, n, n], 原输入 dtype
        残差混合矩阵（未归一化），需输入 Sinkhorn-Knopp 生成双随机矩阵
    """
    M = phi.shape[-1]
    assert M % 32 == 0, "输入的 M 必须是 32 的倍数"

    shape = ops.shape(x_norm)
    B, T = shape[0], shape[1]
    h_native = ops.cast(ops.matmul(x_norm, phi), "float32")

    h_res_raw = alpha_res * h_native[..., : n * n] + bias_res
    h_pre_raw = alpha_pre * h_native[..., n * n : n * (n + 1)] + bias_pre
    h_post_raw = alpha_post * h_native[..., n * (n + 1) : n * (n + 2)] + bias_post

    h_res_reshaped = ops.reshape(h_res_raw, (B, T, n, n))
    return (
        h_pre_raw,
        h_post_raw,
        h_res_reshaped,
    )


def mhc_pre_op_fused(
    x,
    h_res_reshaped,
    h_pre_raw,
    num_iters=20,
    eps=1e-8,
):
    H_res = sinkhorn_knopp(h_res_reshaped, num_iters, eps)
    x_layer_in = stream_aggregate(x, h_pre_raw)
    return x_layer_in, H_res



import os
import torch
import numpy as np

# 设置环境变量
os.environ["CUDA_VISIBLE_DEVICES"] = "1"

# 导入算子
from rwkv_ops.mhc_kernel.native_op import mhc_post_op as native_mhc_op
from rwkv_ops.mhc_kernel.torch_triton_op.mhc_post_op import mhc_post_op as triton_mhc_op

# ------------------------------------------------------------------
# 1. 配置参数与构造输入
# ------------------------------------------------------------------
B, T, n, C = 256, 256, 4, 512
device = "cuda"
dtype = torch.bfloat16

# 创建随机输入
# 为保证公平对比，我们先创建全精度的数据，再转为目标类型
layer_out_raw = torch.randn(B, T, C, device=device).bfloat16()
x_expanded_raw = torch.randn(B, T, n, C, device=device).bfloat16()
h_post_raw_raw = torch.randn(B, T, n, device=device).float()
H_res_raw = torch.randn(B, T, n, n, device=device).float()

# Triton 版本的输入 (BF16/FP32 混合)
# 注意：按照我们的封装，x 和 layer_out 是 BF16，h 和 H 是 FP32
layer_out_triton = layer_out_raw.clone().to(dtype).requires_grad_(True)
x_expanded_triton = x_expanded_raw.clone().to(dtype).requires_grad_(True)
h_post_triton = h_post_raw_raw.clone().requires_grad_(True)
H_res_triton = H_res_raw.clone().requires_grad_(True)

# Native 版本的输入 (克隆一份以防梯度污染)
layer_out_native = layer_out_raw.clone().to(dtype).requires_grad_(True)
x_expanded_native = x_expanded_raw.clone().to(dtype).requires_grad_(True)
h_post_native = h_post_raw_raw.clone().requires_grad_(True)
H_res_native = H_res_raw.clone().requires_grad_(True)


# ------------------------------------------------------------------
# 2. 工具函数：数值比较
# ------------------------------------------------------------------
def test_is_close(name, x1, x2, atol=1e-2, rtol=1e-2):
    # 转为 numpy 比较
    x1_np = x1.detach().cpu().float().numpy()
    x2_np = x2.detach().cpu().float().numpy()

    if np.isnan(x1_np).any() or np.isnan(x2_np).any():
        print(f"❌❌ {name} 存在 NaN ❌❌")
        return

    try:
        np.testing.assert_allclose(x1_np, x2_np, atol=atol, rtol=rtol)
        print(f"✅ {name} 数值一致 (max_diff: {np.abs(x1_np - x2_np).max():.6f})")
    except AssertionError as e:
        print(f"❌ {name} 数值不一致")
        print(e)


# ------------------------------------------------------------------
# 3. 前向测试 (Forward)
# ------------------------------------------------------------------
print("--- 开始前向测试 ---")

# Triton 前向
output_triton = triton_mhc_op(
    layer_out=layer_out_triton,
    x_expanded=x_expanded_triton,
    h_post_raw=h_post_triton,
    H_res=H_res_triton,
)

# Native 前向
output_native = native_mhc_op(
    layer_out=layer_out_native,
    x_expanded=x_expanded_native,
    h_post_raw=h_post_native,
    H_res=H_res_native,
)

test_is_close("Forward Output", output_native, output_triton, atol=1e-2, rtol=1e-2)

# ------------------------------------------------------------------
# 4. 反向测试 (Backward)
# ------------------------------------------------------------------
print("\n--- 开始反向测试 ---")

# 构造一个标量 Loss 进行反向传播
# 使用 mean() 的平方作为一个简单的 loss
loss_triton = (output_triton.float() ** 2).mean()
loss_triton.backward()

loss_native = (output_native.float() ** 2).mean()
loss_native.backward()

# 梯度列表
grad_names = ["layer_out", "x_expanded", "h_post_raw", "H_res"]
triton_grads = [
    layer_out_triton.grad,
    x_expanded_triton.grad,
    h_post_triton.grad,
    H_res_triton.grad,
]
native_grads = [
    layer_out_native.grad,
    x_expanded_native.grad,
    h_post_native.grad,
    H_res_native.grad,
]

for name, g_native, g_triton in zip(grad_names, native_grads, triton_grads):
    if g_native is None or g_triton is None:
        print(f"❌ {name} 梯度不存在")
        continue
    test_is_close(f"Gradient: {name}", g_native, g_triton, atol=1e-2, rtol=1e-2)

print("\n🎉🎉 MHC 全融合算子数值校验结束 🎉🎉")


# ------------------------------------------------------------------
# 5. 性能基准测试 (Benchmark)
# ------------------------------------------------------------------
print("\n" + "=" * 40)
print("🚀 开始性能基准测试 (Speed Benchmark)")
print("=" * 40)

# 定义 Benchmark 配置
n_warmup = 10
n_repeat = 100

# 准备 torch.compile 版本
# mode="reduce-overhead" 适合小算子，"max-autotune" 适合大计算量
try:

    @torch.compile
    def compiled_mhc_op(
        layer_out: torch.Tensor,
        x_expanded: torch.Tensor,
        h_post_raw: torch.Tensor,
        H_res: torch.Tensor,
    ) -> torch.Tensor:
        dtype_out = x_expanded.dtype
        l_f32 = layer_out.float()
        x_f32 = x_expanded.float()
        h_f32 = h_post_raw.float()
        H_f32 = H_res.float()
        w_f32 = torch.sigmoid(h_f32) * 2.0
        x_mixed_f32 = torch.matmul(H_f32, x_f32)
        x_delta_f32 = w_f32.unsqueeze(-1) * l_f32.unsqueeze(-2)
        x_next_f32 = x_mixed_f32 + x_delta_f32
        return x_next_f32.to(dtype_out)
except Exception as e:
    print(f"⚠️ torch.compile 不可用: {e}")
    compiled_mhc_op = None


def benchmark_forward(op_func, name, l, x, h, H):
    # 预热
    for _ in range(n_warmup):
        _ = op_func(l, x, h, H)
    torch.cuda.synchronize()

    # 计时
    start_event = torch.cuda.Event(enable_timing=True)
    end_event = torch.cuda.Event(enable_timing=True)

    start_event.record()
    for _ in range(n_repeat):
        _ = op_func(l, x, h, H)
    end_event.record()

    torch.cuda.synchronize()
    elapsed_time_ms = start_event.elapsed_time(end_event) / n_repeat
    print(f"👉 {name:15s} Forward: {elapsed_time_ms:.4f} ms")
    return elapsed_time_ms


def benchmark_backward(op_func, name, l, x, h, H):
    # 构造 inputs 并设置 grad
    inputs = [t.clone().detach().requires_grad_(True) for t in [l, x, h, H]]

    # 预热
    for _ in range(n_warmup):
        out = op_func(*inputs)
        loss = (out.float() ** 2).mean()
        # 清空梯度
        for t in inputs:
            if t.grad is not None:
                t.grad = None
        loss.backward()
    torch.cuda.synchronize()

    # 计时
    start_event = torch.cuda.Event(enable_timing=True)
    end_event = torch.cuda.Event(enable_timing=True)

    start_event.record()
    for _ in range(n_repeat):
        out = op_func(*inputs)
        loss = (out.float() ** 2).mean()
        # 模拟真实训练：清空梯度 -> 反向传播
        for t in inputs:
            if t.grad is not None:
                t.grad = None
        loss.backward()
    end_event.record()

    torch.cuda.synchronize()
    elapsed_time_ms = start_event.elapsed_time(end_event) / n_repeat
    print(f"👉 {name:15s} Backward (includes Forward): {elapsed_time_ms:.4f} ms")
    return elapsed_time_ms


# ------------------------------------------------------------------
# 运行测试
# ------------------------------------------------------------------
# 使用同样形状的数据进行测试
# 增加一些 Batch Size 模拟真实高负载场景 (可选)
# B_bench, T_bench = 8, 128
B_bench, T_bench = B, T

print(f"配置: B={B_bench}, T={T_bench}, n={n}, C={C}, dtype={dtype}")

l_in = torch.randn(B_bench, T_bench, C, device=device, dtype=dtype)
x_in = torch.randn(B_bench, T_bench, n, C, device=device, dtype=dtype)
h_in = torch.randn(B_bench, T_bench, n, device=device, dtype=torch.float32)
H_in = torch.randn(B_bench, T_bench, n, n, device=device, dtype=torch.float32)

print("\n--- Forward Speed ---")
t_native_fwd = benchmark_forward(
    native_mhc_op, "Native (Eager)", l_in, x_in, h_in, H_in
)
t_triton_fwd = benchmark_forward(triton_mhc_op, "Custom Triton", l_in, x_in, h_in, H_in)
if compiled_mhc_op:
    t_compile_fwd = benchmark_forward(
        compiled_mhc_op, "torch.compile", l_in, x_in, h_in, H_in
    )

print("\n--- Backward Speed (Forward + Backward) ---")
t_native_bwd = benchmark_backward(
    native_mhc_op, "Native (Eager)", l_in, x_in, h_in, H_in
)
t_triton_bwd = benchmark_backward(
    triton_mhc_op, "Custom Triton", l_in, x_in, h_in, H_in
)
if compiled_mhc_op:
    t_compile_bwd = benchmark_backward(
        compiled_mhc_op, "torch.compile", l_in, x_in, h_in, H_in
    )

# ------------------------------------------------------------------
# 总结分析
# ------------------------------------------------------------------
print("\n" + "=" * 40)
print("📊 性能提升总结 (Speedup)")
print("=" * 40)


def print_speedup(base_name, base_time, target_name, target_time):
    speedup = base_time / target_time
    print(f"{target_name} vs {base_name}: {speedup:.2f}x faster")


print(">>> Forward:")
print_speedup("Native", t_native_fwd, "Triton", t_triton_fwd)
if compiled_mhc_op:
    print_speedup("Compile", t_compile_fwd, "Triton", t_triton_fwd)

print("\n>>> Backward:")
print_speedup("Native", t_native_bwd, "Triton", t_triton_bwd)
if compiled_mhc_op:
    print_speedup("Compile", t_compile_bwd, "Triton", t_triton_bwd)

print("\n🎉🎉🎉🎉 所有测试结束 🎉🎉🎉🎉")



# copy right from https://github.com/infiy-quine/RWKV6_Keras_Operator
import os
import keras
from keras import ops
from distutils.util import strtobool
from packaging import version


def get_rwkv6_kernel(KERNEL_TYPE="native"):
    ops_kernel = True
    if KERNEL_TYPE == "cuda":
        if keras.config.backend() == "jax":
            import jax

            if version.parse(jax.__version__) < version.parse("0.6.0"):
                from .jax_rwkv_kernel import RWKVKernelOperator as CudaOperator

                ops_kernel = False
            else:
                CudaOperator = None
        elif keras.config.backend() == "torch":
            from .torch_rwkv_kernel import RWKVKernelOperator as CudaOperator

            ops_kernel = False
        else:
            CudaOperator = None
    else:
        CudaOperator = None
    from .ops_rwkv_kernel import RWKVKernelOperator as OpsOperator

    class RWKVKernelOperator:
        def __init__(self, head_size, max_sequence_length, ops_loop=False):
            self.enbale_cuda = CudaOperator is not None

            if self.enbale_cuda:
                self.cuda_operator = CudaOperator(head_size, max_sequence_length)

            self.ops_operator = OpsOperator(head_size, max_sequence_length)

            self.ops_loop = ops_loop

        def __call__(
            self, r, k, v, w, u, with_state=False, init_state=None, state_map=None
        ):
            seq_len = r.shape[1]

            def call_parallel():
                if self.enbale_cuda:
                    return self.cuda_operator(
                        r=r,
                        k=k,
                        v=v,
                        w=w,
                        u=u,
                        with_state=with_state,
                        init_state=init_state,
                        state_map=state_map,
                    )
                else:
                    return self.ops_operator(
                        r=r,
                        k=k,
                        v=v,
                        w=w,
                        u=u,
                        with_state=with_state,
                        init_state=init_state,
                        state_map=state_map,
                    )

            def call_one_step():
                return self.ops_operator(
                    r=r,
                    k=k,
                    v=v,
                    w=w,
                    u=u,
                    with_state=with_state,
                    init_state=init_state,
                    state_map=state_map,
                )

            if not self.ops_loop:
                return ops.cond(
                    seq_len != 1 and not ops_kernel, call_parallel, call_one_step
                )
            else:
                return call_parallel()

    return RWKVKernelOperator


# from .ops_rwkv_kernal import RWKVKernelOperator as OPSKernelOperator


"""
新增三个参数
return_state 布尔类型 是否返回最终的state,如果想自定义init_state也需要启用这个开关

init_state
    当init_state省缺时，则使用全零初始化BatchSize维度上的状态。
    形状: (state_kinds,num_heads,head_size, head_size)， 其中state_kinds为小于等于Batch_Size的正整数
    精度: 在r为fp16时 init_state为fp32 其余时候类型与r相同


state_map
    形状: (Batch_Size,)
    精度: int64, list[int]
    这个数组定义了state到r上每个Batch维度切片间的映射关系
    取值范围: [0, state_kinds)

返回:
    output, output_state 

def __call__(self,r, k, v, w, u, return_state=False, init_state=None, state_map=None):




"""



import os
import torch
from torch.utils.cpp_extension import load


def get_torch_generalized_delta_rule_single_step(HEAD_SIZE=64):
    flags = [
        "-res-usage",
        f"-D_C_={HEAD_SIZE}",
        "-D_CHUNK_LEN_=1",
        "--use_fast_math",
        "-O3",
        "-Xptxas -O3",
        "--extra-device-vectorization",
    ]
    current_dir = os.path.dirname(os.path.abspath(__file__))
    load(
        name="wind_backstepping_single_step",
        sources=[
            os.path.join(current_dir, "wkv7_single_step_cuda.cu"),
            os.path.join(current_dir, "wkv7_single_step_op.cpp"),
        ],
        is_python_module=False,
        verbose=False,
        extra_cuda_cflags=flags,
    )

    class WindBacksteppingSingleStep(torch.autograd.Function):
        @staticmethod
        def forward(ctx, w, q, k, v, a, b, h0):
            DTYPE = q.dtype
            w = w.contiguous().bfloat16()
            q = q.contiguous().bfloat16()
            k = k.contiguous().bfloat16()
            v = v.contiguous().bfloat16()
            a = a.contiguous().bfloat16()
            b = b.contiguous().bfloat16()
            h0 = h0.contiguous().float()
            y = torch.empty_like(v)
            h1 = torch.empty_like(h0)
            torch.ops.wind_backstepping_single_step.forward_single_step(
                w, q, k, v, a, b, h0, y, h1
            )
            return y.to(DTYPE), h1

        @staticmethod
        def backward(ctx, *grads):
            raise NotImplementedError("single-step kernel does not support backward")

    def run_single_step(w, q, k, v, a, b, h0):
        return WindBacksteppingSingleStep.apply(w, q, k, v, a, b, h0)

    def generalized_delta_rule(
        r: torch.Tensor,
        w: torch.Tensor,
        k: torch.Tensor,
        v: torch.Tensor,
        a: torch.Tensor,
        b: torch.Tensor,
        *,
        initial_state: torch.Tensor = None,
        output_final_state: bool = True,
        head_first: bool = False,
    ):
        """
        单步 RWKV7 前向，输入形状：
          head_first=False  -> (B, 1, H, K)   **默认**
          head_first=True   -> (B, H, 1, K)
        输出形状与输入保持一致。
        """
        if w.device.type != "cuda":
            from ..native_keras_op import generalized_delta_rule

            return generalized_delta_rule(
                r=r,
                k=k,
                v=v,
                a=a,
                b=b,
                w=w,
                initial_state=initial_state,
                output_final_state=output_final_state,
            )
        # 1. 统一先转成 (B, H, K)
        if head_first:  # (B, H, 1, K) -> (B, H, K)
            r = r.squeeze(2)
            w = w.squeeze(2)
            k = k.squeeze(2)
            v = v.squeeze(2)
            a = a.squeeze(2)
            b = b.squeeze(2)
        else:  # (B, 1, H, K) -> (B, H, K)
            r = r.squeeze(1)
            w = w.squeeze(1)
            k = k.squeeze(1)
            v = v.squeeze(1)
            a = a.squeeze(1)
            b = b.squeeze(1)

        B, H, K = r.shape
        if initial_state is None:
            initial_state = torch.zeros(
                B, H, K, K, dtype=torch.float32, device=r.device
            )

        # 2. 计算
        y, h1 = run_single_step(w, r, k, v, a, b, initial_state)  # y:(B,H,K)
        y = y.unsqueeze(1)  # (B, 1, H, K)

        return (y, h1) if output_final_state else y

    return generalized_delta_rule



import os

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["KERAS_BACKEND"] = "torch"
os.environ["KERNEL_TYPE"] = "cuda"

import numpy as np
import torch
from torch.nn import functional as F
from keras import ops

# ------------------------------------------------------------------
# 1. 构造输入
# ------------------------------------------------------------------
T = 1
B = 5
H = 6
K = 64
inputs = [np.random.randn(B, T, H, K) for _ in range(30)]
torch_inputs = [torch.from_numpy(t).bfloat16().cuda() for t in inputs]

a = -F.normalize(torch_inputs[3], dim=-1, p=2.0)
b = F.normalize(torch_inputs[3], dim=-1, p=2.0)

w = torch_inputs[4]  # decay / gate
r = torch_inputs[0]  # receptance
k = torch_inputs[1]
v = torch_inputs[2]
w = -ops.softplus(w) - 0.5
h0 = torch.from_numpy(np.random.randn(B, H, K, K)).float().cuda()
# ------------------------------------------------------------------
# 2. CUDA 版本前向 + 反向
# ------------------------------------------------------------------
from rwkv_ops import rwkv7_op_rnn  # 即 get_generalized_delta_rule 返回的函数

# 要求计算梯度，必须设置 requires_grad
for t in [r, k, v, a, b, w, h0]:
    t.requires_grad_(True)

cuda_out, cuda_state = rwkv7_op_rnn(
    r=r, k=k, v=v, a=a, b=b, w=w, initial_state=h0, output_final_state=True
)


# ------------------------------------------------------------------
# 3. Native 版本前向 + 反向
# ------------------------------------------------------------------
from rwkv_ops.rwkv7_kernel.native_keras_op import generalized_delta_rule

# 重新构造一份 detach 的新张量，避免梯度混淆
r_n = r.detach().clone().requires_grad_(True)
k_n = k.detach().clone().requires_grad_(True)
v_n = v.detach().clone().requires_grad_(True)
a_n = a.detach().clone().requires_grad_(True)
b_n = b.detach().clone().requires_grad_(True)
w_n = w.detach().clone().requires_grad_(True)
h0_n = h0.detach().clone().requires_grad_(True)

native_out, native_state = generalized_delta_rule(
    r=r_n, k=k_n, v=v_n, a=a_n, b=b_n, w=w_n, initial_state=h0_n
)


# ------------------------------------------------------------------
# 4. 前向结果比较
# ------------------------------------------------------------------
np.testing.assert_allclose(
    ops.convert_to_numpy(native_out.float()),
    ops.convert_to_numpy(cuda_out.float()),
    atol=1e-3,
    rtol=1e-2,
)
np.testing.assert_allclose(
    ops.convert_to_numpy(native_state.float()),
    ops.convert_to_numpy(cuda_state.float()),
    atol=1e-5,
    rtol=1e-5,
)
print("✅ 前向输出一致")
print("🎉🎉🎉🎉test_script/test_torch_cuda_kernel_single.py测试结束🎉🎉🎉🎉")



import triton
import triton.language as tl


@triton.autotune(
    configs=[
        triton.Config(
            {"BLOCK_CHANNEL": block_c}, num_warps=num_warps, num_stages=num_stages
        )
        for block_c in [128, 256, 512, 1024]
        for num_warps in [4, 8]
        for num_stages in [2, 3, 4]
    ],
    key=["CHANNEL_SIZE"],
)
@triton.jit
def mhc_fused_forward_kernel(
    # --- 指针参数 ---
    x_expanded_ptr,
    h_post_raw_ptr,
    H_res_ptr,
    layer_out_ptr,
    output_ptr,
    # --- 步幅参数 (tl.constexpr) ---
    stride_output_batch_time: tl.constexpr,
    stride_output_n_size: tl.constexpr,
    stride_output_channel: tl.constexpr,
    stride_x_batch_time: tl.constexpr,
    stride_x_n_size: tl.constexpr,
    stride_x_channel: tl.constexpr,
    stride_h_batch_time: tl.constexpr,
    stride_h_n_size: tl.constexpr,
    stride_H_batch_time: tl.constexpr,
    stride_H_n_size_1: tl.constexpr,
    stride_H_n_size_2: tl.constexpr,
    stride_layer_out_batch_time: tl.constexpr,
    stride_layer_out_channel: tl.constexpr,
    # --- 编译时常量 ---
    CHANNEL_SIZE: tl.constexpr,
    NSIZE: tl.constexpr,
    BLOCK_CHANNEL: tl.constexpr,
):
    pid_batch_time = tl.program_id(0)
    pid_channel_block = tl.program_id(1)

    offset_channel = pid_channel_block * BLOCK_CHANNEL + tl.arange(0, BLOCK_CHANNEL)
    mask_channel = offset_channel < CHANNEL_SIZE

    # 基础指针
    x_expanded_base = x_expanded_ptr + pid_batch_time * stride_x_batch_time
    h_post_raw_base = h_post_raw_ptr + pid_batch_time * stride_h_batch_time
    H_res_base = H_res_ptr + pid_batch_time * stride_H_batch_time
    layer_out_base = layer_out_ptr + pid_batch_time * stride_layer_out_batch_time
    output_base = output_ptr + pid_batch_time * stride_output_batch_time

    # 索引 n: [0, 1, ..., n-1]
    index_n = tl.arange(0, NSIZE)

    # 一次性读取 h_post [NSIZE]
    h_post_values = tl.load(h_post_raw_base + index_n * stride_h_n_size).to(tl.float32)
    weight_values = tl.sigmoid(h_post_values) * 2.0

    layer_out_values = tl.load(
        layer_out_base + offset_channel * stride_layer_out_channel,
        mask=mask_channel,
        other=0.0,
    ).to(tl.float32)

    # accumulator = weight * layer_out
    accumulator = weight_values[:, None] * layer_out_values[None, :]

    for k in tl.static_range(NSIZE):
        # A. 加载 x_expanded 的第 k 行 [1, BLOCK_CHANNEL]
        x_row_k_ptr = (
            x_expanded_base + k * stride_x_n_size + offset_channel * stride_x_channel
        )
        x_row_k_value = tl.load(x_row_k_ptr, mask=mask_channel, other=0.0).to(
            tl.float32
        )

        H_column_k_ptr = (
            H_res_base + index_n * stride_H_n_size_1 + k * stride_H_n_size_2
        )
        H_column_k_value = tl.load(H_column_k_ptr).to(tl.float32)

        # C. 累加外积
        # [N, 1] * [1, BLOCK_C] -> [N, BLOCK_C]
        accumulator += H_column_k_value[:, None] * x_row_k_value[None, :]

    index_n_output = index_n[:, None]
    index_c_output = offset_channel[None, :]

    output_target_ptrs = (
        output_base
        + index_n_output * stride_output_n_size
        + index_c_output * stride_output_channel
    )

    tl.store(
        output_target_ptrs, accumulator.to(tl.bfloat16), mask=mask_channel[None, :]
    )


@triton.autotune(
    configs=[
        triton.Config(
            {"BLOCK_CHANNEL": block_c}, num_warps=num_warps, num_stages=num_stages
        )
        for block_c in [128, 256, 512, 1024]
        for num_warps in [2, 4, 8]
        for num_stages in [2, 3, 4]
    ],
    key=["CHANNEL_SIZE"],
)
@triton.jit
def mhc_fused_backward_kernel(
    # --- 指针 ---
    x_ptr,
    h_ptr,
    H_ptr,
    l_ptr,
    g_ptr,
    gx_ptr,
    gh_ptr,
    gH_ptr,
    gl_ptr,
    # --- 步幅 ---
    stride_x_bt: tl.constexpr,
    stride_x_n: tl.constexpr,
    stride_x_c: tl.constexpr,
    stride_h_bt: tl.constexpr,
    stride_h_n: tl.constexpr,
    stride_H_bt: tl.constexpr,
    stride_H_n1: tl.constexpr,
    stride_H_n2: tl.constexpr,
    stride_l_bt: tl.constexpr,
    stride_l_c: tl.constexpr,
    stride_g_bt: tl.constexpr,
    stride_g_n: tl.constexpr,
    stride_g_c: tl.constexpr,
    stride_gx_bt: tl.constexpr,
    stride_gx_n: tl.constexpr,
    stride_gx_c: tl.constexpr,
    stride_gl_bt: tl.constexpr,
    stride_gl_c: tl.constexpr,
    stride_gh_bt: tl.constexpr,
    stride_gh_n: tl.constexpr,
    stride_gH_bt: tl.constexpr,
    stride_gH_n1: tl.constexpr,
    stride_gH_n2: tl.constexpr,
    # --- 常量 ---
    CHANNEL_SIZE: tl.constexpr,
    NSIZE: tl.constexpr,
    BLOCK_CHANNEL: tl.constexpr,
):
    # 1. 获取 Program ID (仅处理 Batch * Time 维度)
    pid_bt = tl.program_id(0)

    # 2. 构造 N 维度的索引 (静态长度)
    off_n = tl.arange(0, NSIZE)
    mask_H = (off_n[:, None] < NSIZE) & (off_n[None, :] < NSIZE)

    # 3. 基础指针偏移
    p_h = h_ptr + pid_bt * stride_h_bt + off_n * stride_h_n
    p_H = (
        H_ptr
        + pid_bt * stride_H_bt
        + (off_n[:, None] * stride_H_n1 + off_n[None, :] * stride_H_n2)
    )

    # 4. 预加载 H 矩阵 [N, N] 和 h 激活值 [N]
    # 这些在处理整行 C 的过程中是常数，存在寄存器里
    h_vals = tl.load(p_h).to(tl.float32)
    sig_h = tl.sigmoid(h_vals)
    w_vals = sig_h * 2.0
    dw_vals = sig_h * (1.0 - sig_h) * 2.0

    H_vals = tl.load(p_H, mask=mask_H, other=0.0).to(tl.float32)

    # 5. 初始化规约累加器 (Persistent Registers)
    acc_gh = tl.zeros([NSIZE], dtype=tl.float32)
    acc_gH = tl.zeros([NSIZE, NSIZE], dtype=tl.float32)

    # -----------------------------------------------------------
    # 6. 主循环：跨步遍历 Channel 维度
    # -----------------------------------------------------------
    for start_c in tl.static_range(0, CHANNEL_SIZE, BLOCK_CHANNEL):
        off_c = start_c + tl.arange(0, BLOCK_CHANNEL)
        mask_c = off_c < CHANNEL_SIZE
        mask_2d = (off_n[:, None] < NSIZE) & (off_c[None, :] < CHANNEL_SIZE)

        # 加载分块数据
        # l_chunk: [C]
        l_chunk = tl.load(
            l_ptr + pid_bt * stride_l_bt + off_c * stride_l_c, mask=mask_c, other=0.0
        ).to(tl.float32)
        # x_chunk: [N, C]
        x_chunk = tl.load(
            x_ptr
            + pid_bt * stride_x_bt
            + (off_n[:, None] * stride_x_n + off_c[None, :] * stride_x_c),
            mask=mask_2d,
            other=0.0,
        ).to(tl.float32)
        # g_chunk: [N, C]
        g_chunk = tl.load(
            g_ptr
            + pid_bt * stride_g_bt
            + (off_n[:, None] * stride_g_n + off_c[None, :] * stride_g_c),
            mask=mask_2d,
            other=0.0,
        ).to(tl.float32)

        # --- Task A: grad_layer_out [C] ---
        # 结果 = sum_n (g[n, c] * w[n])
        gl_chunk = tl.sum(g_chunk * w_vals[:, None], axis=0)
        tl.store(
            gl_ptr + pid_bt * stride_gl_bt + off_c * stride_gl_c,
            gl_chunk.to(tl.bfloat16),
            mask=mask_c,
        )

        # --- Task B: grad_x [N, C] ---
        # 结果 = H^T @ g -> [N, N] @ [N, C]
        # 使用广播计算：H[i, j, 1] * g[i, 1, c] -> [i, j, c] -> sum over i
        gx_chunk = tl.sum(H_vals[:, :, None] * g_chunk[:, None, :], axis=0)
        tl.store(
            gx_ptr
            + pid_bt * stride_gx_bt
            + (off_n[:, None] * stride_gx_n + off_c[None, :] * stride_gx_c),
            gx_chunk.to(tl.bfloat16),
            mask=mask_2d,
        )

        # --- Task C: 累加 grad_h 分量 ---
        # 局部规约：sum_c (g[n, c] * l[c])
        acc_gh += tl.sum(g_chunk * l_chunk[None, :], axis=1) * dw_vals

        # --- Task D: 累加 grad_H 分量 ---
        # 局部规约：sum_c (g[i, c] * x[j, c])
        # 使用广播：g[i, 1, c] * x[1, j, c] -> [i, j, c] -> sum over c
        acc_gH += tl.sum(g_chunk[:, None, :] * x_chunk[None, :, :], axis=2)

    # -----------------------------------------------------------
    # 7. 循环结束：写回规约完成的 dh 和 dH
    # -----------------------------------------------------------
    tl.store(
        gh_ptr + pid_bt * stride_gh_bt + off_n * stride_gh_n, acc_gh.to(tl.float32)
    )
    tl.store(
        gH_ptr
        + pid_bt * stride_gH_bt
        + (off_n[:, None] * stride_gH_n1 + off_n[None, :] * stride_gH_n2),
        acc_gH.to(tl.float32),
        mask=mask_H,
    )



import triton
import triton.language as tl


@triton.autotune(
    configs=[
        triton.Config(
            {"BLOCK_BT": bt_num, "BLOCK_C": csize},
            num_warps=num_warps,
            num_stages=num_stages,
        )
        for bt_num in [16, 32]
        for csize in [128, 256, 512]
        for num_warps in [4, 8]
        for num_stages in [2, 3, 4]
    ],
    key=["CSIZE", "Total_BT_CONST"],
)
@triton.jit
def sinkhorn_aggregate_fused_kernel(
    # --- 1. 指针参数 ---
    x_ptr,
    h_res_in_ptr,
    h_pre_in_ptr,
    out_ptr,
    H_res_out_ptr,
    # --- 2. 编译时常量 ---
    Total_BT_CONST: tl.constexpr,
    NSIZE: tl.constexpr,
    CSIZE: tl.constexpr,
    NUM_ITERS: tl.constexpr,
    EPS: tl.constexpr,
    # --- 3. 步幅参数 (constexpr) ---
    stride_x_bt: tl.constexpr,
    stride_x_n: tl.constexpr,
    stride_x_c: tl.constexpr,
    stride_h_res_in_bt: tl.constexpr,
    stride_h_res_in_n1: tl.constexpr,
    stride_h_res_in_n2: tl.constexpr,
    stride_h_pre_in_bt: tl.constexpr,
    stride_h_pre_in_n: tl.constexpr,
    stride_out_bt: tl.constexpr,
    stride_out_c: tl.constexpr,
    stride_Hr_out_bt: tl.constexpr,
    stride_Hr_out_n1: tl.constexpr,
    stride_Hr_out_n2: tl.constexpr,
    # --- 4. 调优参数 ---
    BLOCK_BT: tl.constexpr,
    BLOCK_C: tl.constexpr,
):
    pid_bt = tl.program_id(0)
    pid_c = tl.program_id(1)

    offs_bt = pid_bt * BLOCK_BT + tl.arange(0, BLOCK_BT)
    mask_bt = offs_bt < Total_BT_CONST

    # -----------------------------------------------------------
    # PART 1: Sinkhorn-Knopp (仅 pid_c == 0 执行)
    # -----------------------------------------------------------
    if pid_c == 0:
        offs_n1 = tl.arange(0, NSIZE)
        offs_n2 = tl.arange(0, NSIZE)

        h_res_ptr_loc = (
            h_res_in_ptr
            + offs_bt[:, None, None] * stride_h_res_in_bt
            + offs_n1[None, :, None] * stride_h_res_in_n1
            + offs_n2[None, None, :] * stride_h_res_in_n2
        )

        h_res = tl.load(h_res_ptr_loc, mask=mask_bt[:, None, None], other=0.0).to(
            tl.float32
        )

        max_val = tl.max(h_res, axis=2)
        max_val = tl.max(max_val, axis=1)
        h_res_stabilized = h_res - max_val[:, None, None]
        P = tl.exp(h_res_stabilized)

        for _ in tl.static_range(NUM_ITERS):
            row_sum = tl.sum(P, axis=2)
            P = P / (row_sum[:, :, None] + EPS)
            col_sum = tl.sum(P, axis=1)
            P = P / (col_sum[:, None, :] + EPS)

        H_out_loc = (
            H_res_out_ptr
            + offs_bt[:, None, None] * stride_Hr_out_bt
            + offs_n1[None, :, None] * stride_Hr_out_n1
            + offs_n2[None, None, :] * stride_Hr_out_n2
        )
        tl.store(H_out_loc, P, mask=mask_bt[:, None, None])

    # -----------------------------------------------------------
    # PART 2: Stream Aggregate
    # -----------------------------------------------------------

    # 2.1 加载聚合权重 (一次性加载所有流的权重，效率最高)
    offs_n = tl.arange(0, NSIZE)
    h_pre_ptr_loc = (
        h_pre_in_ptr
        + offs_bt[:, None] * stride_h_pre_in_bt
        + offs_n[None, :] * stride_h_pre_in_n
    )

    h_pre = tl.load(h_pre_ptr_loc, mask=mask_bt[:, None], other=0.0).to(tl.float32)
    w_pre = tl.sigmoid(h_pre)  # [BLOCK_BT, N]

    # 2.2 聚合计算
    offs_c = pid_c * BLOCK_C + tl.arange(0, BLOCK_C)
    mask_c = offs_c < CSIZE

    acc = tl.zeros([BLOCK_BT, BLOCK_C], dtype=tl.float32)

    # 用于列提取的辅助索引 [1, N]
    n_range = tl.arange(0, NSIZE)[None, :]

    for k in tl.static_range(NSIZE):
        # --- 关键修改: 使用 Masking 提取第 k 列 ---
        # w_pre: [BLOCK_BT, N]
        # col_mask: [1, N] (例如 k=2 -> [0,0,1,0])
        col_mask = n_range == k

        # 1. 广播乘法: 保留第 k 列，其他置零
        # 2. Sum(axis=1): 降维成 [BLOCK_BT]
        # 3. [:, None]: 广播成 [BLOCK_BT, 1] 用于后续乘法
        w_k = tl.sum(w_pre * col_mask, axis=1)[:, None]

        # 加载 x
        x_ptr_k = (
            x_ptr
            + offs_bt[:, None] * stride_x_bt
            + k * stride_x_n
            + offs_c[None, :] * stride_x_c
        )

        load_mask = mask_bt[:, None] & mask_c[None, :]
        val_x = tl.load(x_ptr_k, mask=load_mask, other=0.0).to(tl.float32)

        acc += val_x * w_k

    # 2.3 写回
    out_loc = (
        out_ptr + offs_bt[:, None] * stride_out_bt + offs_c[None, :] * stride_out_c
    )

    store_mask = mask_bt[:, None] & mask_c[None, :]
    tl.store(out_loc, acc.to(tl.bfloat16), mask=store_mask)


@triton.autotune(
    configs=[
        triton.Config(
            {"BLOCK_CHANNEL": bc},
            num_warps=num_warps,
            num_stages=num_stages,
        )
        for bc in [128, 256, 512]
        for num_warps in [4, 8]
        for num_stages in [2, 3, 4]
    ],
    key=["CHANNEL_SIZE", "TOTAL_BT_CONST"],
)
@triton.jit
def sinkhorn_aggregate_bwd_kernel(
    # --- 1. 输入指针 ---
    grad_out_ptr,  # [Total_BT, C]
    grad_H_res_out_ptr,  # [Total_BT, n, n]
    x_ptr,  # [Total_BT, n, C]
    h_res_in_ptr,  # [Total_BT, n, n]
    h_pre_in_ptr,  # [Total_BT, n]
    # --- 2. 输出指针 ---
    grad_x_ptr,  # [Total_BT, n, C]
    grad_h_res_in_ptr,  # [Total_BT, n, n]
    grad_h_pre_in_ptr,  # [Total_BT, n]
    # --- 3. 编译时常量 ---
    TOTAL_BT_CONST: tl.constexpr,
    NSIZE: tl.constexpr,
    CHANNEL_SIZE: tl.constexpr,
    NUM_ITERS: tl.constexpr,
    EPS: tl.constexpr,
    # --- 4. 步幅参数 (constexpr) ---
    stride_gout_bt: tl.constexpr,
    stride_gout_c: tl.constexpr,
    stride_gH_bt: tl.constexpr,
    stride_gH_n1: tl.constexpr,
    stride_gH_n2: tl.constexpr,
    stride_x_bt: tl.constexpr,
    stride_x_n: tl.constexpr,
    stride_x_c: tl.constexpr,
    stride_h_res_bt: tl.constexpr,
    stride_h_res_n1: tl.constexpr,
    stride_h_res_n2: tl.constexpr,
    stride_h_pre_bt: tl.constexpr,
    stride_h_pre_n: tl.constexpr,
    stride_gx_bt: tl.constexpr,
    stride_gx_n: tl.constexpr,
    stride_gx_c: tl.constexpr,
    stride_gh_res_bt: tl.constexpr,
    stride_gh_res_n1: tl.constexpr,
    stride_gh_res_n2: tl.constexpr,
    stride_gh_pre_bt: tl.constexpr,
    stride_gh_pre_n: tl.constexpr,
    # --- 5. 调优参数 ---
    BLOCK_CHANNEL: tl.constexpr,
):
    # 强制单 Block 处理整行 C 以消除 atomic_add
    pid_bt = tl.program_id(0)

    # ===========================================================
    # PART 1: 精确 Sinkhorn 梯度 (先处理并写回，释放寄存器)
    # ===========================================================
    off_n1 = tl.arange(0, NSIZE)
    off_n2 = tl.arange(0, NSIZE)

    # [1.1] 加载原始输入 h_res_in 并重算前向 P
    h_res_ptr = (
        h_res_in_ptr
        + pid_bt * stride_h_res_bt
        + off_n1[:, None] * stride_h_res_n1
        + off_n2[None, :] * stride_h_res_n2
    )
    h_res = tl.load(h_res_ptr).to(tl.float32)

    # 前向重算 (指数空间)
    max_val = tl.max(tl.max(h_res, 1), 0)
    P = tl.exp(h_res - max_val)
    for _ in tl.range(NUM_ITERS):
        P /= tl.sum(P, axis=1)[:, None] + EPS
        P /= tl.sum(P, axis=0)[None, :] + EPS

    # [1.2] 加载输出梯度并执行精确 VJP 逆向迭代
    gH_ptr = (
        grad_H_res_out_ptr
        + pid_bt * stride_gH_bt
        + off_n1[:, None] * stride_gH_n1
        + off_n2[None, :] * stride_gH_n2
    )
    dP = tl.load(gH_ptr).to(tl.float32)

    for _ in tl.static_range(NUM_ITERS):
        # 逆向列归一化: dX = dY - Y * sum(dY * Y)
        dP = dP - P * tl.sum(dP * P, axis=0)[None, :]
        # 逆向行归一化
        dP = dP - P * tl.sum(dP * P, axis=1)[:, None]

    # 写回 grad_h_res = dP * P (映射回 Log 空间)
    grad_h_res = dP * P
    gh_res_out_ptr = (
        grad_h_res_in_ptr
        + pid_bt * stride_gh_res_bt
        + off_n1[:, None] * stride_gh_res_n1
        + off_n2[None, :] * stride_gh_res_n2
    )
    tl.store(gh_res_out_ptr, grad_h_res)

    # ===========================================================
    # PART 2: Aggregate 梯度 (Persistent Reduction)
    # ===========================================================

    # [2.1] 准备聚合权重与掩码
    off_n = tl.arange(0, NSIZE)
    n_range = off_n[None, :]
    h_pre_ptr = h_pre_in_ptr + pid_bt * stride_h_pre_bt + off_n * stride_h_pre_n
    h_pre = tl.load(h_pre_ptr).to(tl.float32)
    w_pre = tl.sigmoid(h_pre)
    dw_pre = w_pre * (1.0 - w_pre)  # Sigmoid 导数

    # 持久化累加器
    acc_gh_pre = tl.zeros([NSIZE], dtype=tl.float32)

    # [2.2] 跨步遍历 Channel 维度
    for start_c in tl.static_range(0, CHANNEL_SIZE, BLOCK_CHANNEL):
        off_c = start_c + tl.arange(0, BLOCK_CHANNEL)
        mask_c = off_c < CHANNEL_SIZE

        # 加载 grad_out 块
        gout_ptr = grad_out_ptr + pid_bt * stride_gout_bt + off_c * stride_gout_c
        g_out = tl.load(gout_ptr, mask=mask_c, other=0.0).to(tl.float32)

        # 遍历流计算 grad_x 并规约 grad_h_pre
        for k in tl.static_range(NSIZE):
            # 获取当前流权重 (Masking trick)
            w_k = tl.sum(w_pre * (off_n == k), axis=0)

            # A. 计算并存储 grad_x [k, c]
            gx_chunk = g_out * w_k
            gx_out_ptr = (
                grad_x_ptr
                + pid_bt * stride_gx_bt
                + k * stride_gx_n
                + off_c * stride_gx_c
            )
            tl.store(gx_out_ptr, gx_chunk.to(tl.bfloat16), mask=mask_c)

            # B. 计算 grad_h_pre 规约分量
            x_ptr_k = x_ptr + pid_bt * stride_x_bt + k * stride_x_n + off_c * stride_x_c
            x_chunk = tl.load(x_ptr_k, mask=mask_c, other=0.0).to(tl.float32)

            # dot_sum = sum_c(grad_out * x)
            dot_sum = tl.sum(g_out * x_chunk, axis=0)
            acc_gh_pre += dot_sum * (off_n == k)

    # [2.3] 规约结束，应用 Sigmoid 导数并写回
    final_gh_pre = acc_gh_pre * dw_pre
    gh_pre_out_ptr = (
        grad_h_pre_in_ptr + pid_bt * stride_gh_pre_bt + off_n * stride_gh_pre_n
    )
    tl.store(gh_pre_out_ptr, final_gh_pre)



import os

os.environ["CUDA_VISIBLE_DEVICES"] = "2"
os.environ["KERAS_BACKEND"] = "torch"
os.environ["KERNEL_TYPE"] = "cuda"

import numpy as np
import torch
from torch.nn import functional as F
from keras import ops

# ------------------------------------------------------------------
# 1. 构造输入
# ------------------------------------------------------------------
T = 1024
B = 5
H = 6
K = 64
inputs = [np.random.randn(B, T, H, K) for _ in range(30)]
torch_inputs = [torch.from_numpy(t).bfloat16().cuda() for t in inputs]

a = -F.normalize(torch_inputs[3], dim=-1, p=2.0)
b = F.normalize(torch_inputs[3], dim=-1, p=2.0)

w = torch_inputs[4]  # decay / gate
r = torch_inputs[0]  # receptance
k = torch_inputs[1]
v = torch_inputs[2]
w = -ops.softplus(w) - 0.5
h0 = torch.from_numpy(np.random.randn(B, H, K, K)).float().cuda()

# ------------------------------------------------------------------
# 2. CUDA 版本前向（无 Mask）
# ------------------------------------------------------------------
from rwkv_ops import rwkv7_op_inference  # 推理专用接口

cuda_out, cuda_state = rwkv7_op_inference(
    r=r, k=k, v=v, a=a, b=b, w=w, initial_state=h0, output_final_state=True
)

# ------------------------------------------------------------------
# 3. Native 版本前向（无 Mask）
# ------------------------------------------------------------------
from rwkv_ops.rwkv7_kernel.native_keras_op import generalized_delta_rule

native_out, native_state = generalized_delta_rule(
    r=r, k=k, v=v, a=a, b=b, w=w, initial_state=h0
)


# ------------------------------------------------------------------
# 4. 前向结果比较工具
# ------------------------------------------------------------------
def test_is_close(name, x1, x2, atol=5e-3, rtol=1e-3):
    x1 = ops.convert_to_numpy(ops.cast(x1, "float32"))
    x2 = ops.convert_to_numpy(ops.cast(x2, "float32"))

    # NaN 检查
    if np.isnan(x1).sum() == 0 and np.isnan(x2).sum() == 0:
        print(f"✅ {name} 无 NaN")
    else:
        print(
            f"❌ {name} 存在 NaN! (CUDA: {np.isnan(x1).sum()}, Native: {np.isnan(x2).sum()})"
        )
        return False

    # 完全一致检查（快速路径）
    if np.abs(x1 - x2).max() < 1e-6:
        print(f"✅ {name} 数值完全一致 (Max Diff < 1e-6)")
        return True

    # 阈值检查
    try:
        np.testing.assert_allclose(x1, x2, atol=atol, rtol=rtol)
        max_diff = np.abs(x1 - x2).max()
        print(f"✅ {name} 一致 (Max Diff: {max_diff:.6e})")
        return True
    except AssertionError as e:
        max_diff = np.abs(x1 - x2).max()
        print(f"❌ {name} 不一致! Max Diff: {max_diff:.6e}")
        print(str(e)[:200] + "...")
        return False


# ------------------------------------------------------------------
# 5. 无 Mask 基础测试
# ------------------------------------------------------------------
print("=" * 60)
print("无 Mask 推理测试")
print("=" * 60)
test_is_close("fwd_pred", native_out, cuda_out, atol=1e-5, rtol=1e-2)
test_is_close("fwd_state", native_state, cuda_state, atol=1e-5, rtol=1e-3)
print("🎉 无 Mask 测试完成\n")

# ------------------------------------------------------------------
# 6. 带 Mask 推理测试
# ------------------------------------------------------------------
print("=" * 60)
print("带 Mask 推理测试")
print("=" * 60)

# 构造随机 Mask（30% 冻结）
mask = torch.ones(B, T, device="cuda", dtype=torch.float32)
freeze_indices = torch.rand(B, T) < 0.3
mask[freeze_indices] = 0.0
print(f"Mask 冻结比例: {(mask == 0).float().mean().item():.2%}")

# CUDA 带 Mask 推理
cuda_out_mask, cuda_state_mask = rwkv7_op_inference(
    r=r,
    k=k,
    v=v,
    a=a,
    b=b,
    w=w,
    mask=mask,  # 传入 mask
    initial_state=h0,
    output_final_state=True,
)

# Native 带 Mask 推理
native_out_mask, native_state_mask = generalized_delta_rule(
    r=r,
    k=k,
    v=v,
    a=a,
    b=b,
    w=w,
    mask=mask,  # 传入 mask
    initial_state=h0,
)

# 对比结果
print("\n>>> 随机 Mask 前向对比")
test_is_close("fwd_pred_mask", native_out_mask, cuda_out_mask, atol=1e-5, rtol=1e-2)
test_is_close(
    "fwd_state_mask", native_state_mask, cuda_state_mask, atol=1e-5, rtol=1e-3
)

# ------------------------------------------------------------------
# 7. Mask 物理一致性验证
# ------------------------------------------------------------------
print("\n" + "=" * 60)
print("Mask 物理一致性验证")
print("=" * 60)

# 测试 1: 全 0 Mask，状态应该完全不变
print("\n[测试 1] 全 0 Mask（状态冻结）")
mask_all_zero = torch.zeros(B, T, device="cuda", dtype=torch.float32)
h0_frozen = h0.clone()

out_frozen, state_frozen = rwkv7_op_inference(
    r=r,
    k=k,
    v=v,
    a=a,
    b=b,
    w=w,
    mask=mask_all_zero,
    initial_state=h0_frozen,
    output_final_state=True,
)

diff = (state_frozen - h0_frozen).abs().max().item()
if diff < 1e-5:
    print(f"✅ 状态冻结验证通过 (Max Diff: {diff:.2e})")
else:
    print(f"❌ 状态被错误改变 (Max Diff: {diff:.2e})")

# 测试 2: 全 1 Mask，应与无 Mask 版本完全一致
print("\n[测试 2] 全 1 Mask（与无 Mask 等价）")
mask_all_one = torch.ones(B, T, device="cuda", dtype=torch.float32)

out_one, state_one = rwkv7_op_inference(
    r=r,
    k=k,
    v=v,
    a=a,
    b=b,
    w=w,
    mask=mask_all_one,
    initial_state=h0,
    output_final_state=True,
)

pred_diff = (out_one - cuda_out).abs().max().item()
state_diff = (state_one - cuda_state).abs().max().item()

if pred_diff < 1e-5 and state_diff < 1e-5:
    print("✅ 全 1 Mask 与无 Mask 等价")
    print(f"   输出差异: {pred_diff:.2e}, 状态差异: {state_diff:.2e}")
else:
    print(f"⚠️ 存在微小差异 (输出: {pred_diff:.2e}, 状态: {state_diff:.2e})")

# 测试 3: 边界情况 - 只在最后一个时间步更新
print("\n[测试 3] 仅最后一帧更新（前 T-1 帧冻结）")
mask_last = torch.zeros(B, T, device="cuda", dtype=torch.float32)
mask_last[:, -1] = 1.0  # 只有最后一步更新

out_last, state_last = rwkv7_op_inference(
    r=r,
    k=k,
    v=v,
    a=a,
    b=b,
    w=w,
    mask=mask_last,
    initial_state=h0,
    output_final_state=True,
)

# 验证：最终状态应该只包含最后一步的信息（与 h0 差异不应为零，但也不应过大）
state_change = (state_last - h0).abs().mean().item()
print(f"   状态平均变化量: {state_change:.4f} (应 > 0 且合理)")

print("\n🎉🎉🎉 全部推理测试完成 🎉🎉🎉")



import argparse
import os
import numpy as np


def run_test(backend: str, kernel_type: str):
    print(f"\nTesting RWKV6 with KERAS_BACKEND={backend}, KERNEL_TYPE={kernel_type}")

    # 设置环境变量
    os.environ["KERAS_BACKEND"] = backend
    os.environ["KERNEL_TYPE"] = kernel_type
    from keras import ops

    from rwkv_ops import RWKV6_OP

    B, T, H, C = 2, 16, 6, 64
    r = ops.array(np.random.randn(B, T, H * C), dtype="bfloat16")
    w = ops.array(np.random.randn(B, T, H * C), dtype="bfloat16")
    k = ops.array(np.random.randn(B, T, H * C), dtype="bfloat16")
    v = ops.array(np.random.randn(B, T, H * C), dtype="bfloat16")
    u = ops.array(np.random.randn(1, 1, H * C), dtype="bfloat16")
    rwkv6_op = RWKV6_OP(C, T)
    output, state = rwkv6_op(r, k, v, w, u)
    try:
        output, state = rwkv6_op(r, k, v, w, u)
        print("Output shape:", output.shape)
        assert output.shape == (
            B,
            T,
            H * C,
        ), f"Expected output shape {(B, T, H * C)}, got {output.shape}"
        print(
            "✅ Test passed at %s Backend and %s impplementation"
            % (backend, kernel_type)
        )
    except Exception as e:
        print(f"❌ Test failed: {e}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Test RWKV6 Op with different backends and kernel types."
    )
    parser.add_argument(
        "--backend",
        type=str,
        choices=["torch", "jax", "numpy", "tensorflow", "openvino"],
        required=True,
    )
    parser.add_argument(
        "--kernel-type",
        type=str,
        choices=["triton", "cuda", "native"],
        required=True,
        help="Kernel type to use (triton, cuda, native)",
    )

    args = parser.parse_args()

    run_test(args.backend, args.kernel_type)
    print("🎉🎉🎉🎉test_script/test_rwkv6_kernel.py测试结束🎉🎉🎉🎉")



import os

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["KERAS_BACKEND"] = "jax"
os.environ["KERNEL_TYPE"] = "cuda"

import numpy as np
import jax.numpy as jnp
from keras import ops

# ------------------------------------------------------------------
# 1. 构造输入
# ------------------------------------------------------------------
T = 128  # 可以根据需要调整，推理可以用较大值
B = 5
H = 6
K = 64
inputs = [np.random.randn(B, T, H, K) for _ in range(30)]
jax_inputs = [jnp.asarray(t, "bfloat16") for t in inputs]


def normalize(z, p=2, dim=-1, eps: float = 1e-12):
    # F.normalize like api
    denom = ops.norm(z, ord=p, axis=dim, keepdims=True)
    denom = ops.maximum(denom, eps)
    return z / denom


a = -normalize(jax_inputs[3], dim=-1, p=2.0)
b = normalize(jax_inputs[3], dim=-1, p=2.0)

w = jax_inputs[4]  # decay / gate
r = jax_inputs[0]  # receptance
k = jax_inputs[1]
v = jax_inputs[2]
w = -ops.softplus(w) - 0.5
h0 = jnp.asarray(np.random.randn(B, H, K, K), "float32")


# ------------------------------------------------------------------
# 2. 辅助函数
# ------------------------------------------------------------------
def test_is_close(name, x1, x2, atol=5e-3, rtol=1e-3):
    x1 = ops.convert_to_numpy(ops.cast(x1, "float32"))
    x2 = ops.convert_to_numpy(ops.cast(x2, "float32"))

    # NaN 检查
    if np.sum(np.isnan(x1)) == 0 and np.sum(np.isnan(x2)) == 0:
        print(f"✅ {name} 无 NaN")
    else:
        print(f"❌ {name} 存在 NaN!")
        print(f"   x1 NaN 数量: {np.sum(np.isnan(x1))}")
        print(f"   x2 NaN 数量: {np.sum(np.isnan(x2))}")
        return False

    # 完全一致检查（快速路径）
    if np.allclose(x1, x2, atol=1e-6):
        print(f"✅ {name} 数值完全一致")
        return True

    # 阈值检查
    try:
        np.testing.assert_allclose(x1, x2, atol=atol, rtol=rtol)
        max_diff = np.abs(x1 - x2).max()
        print(f"✅ {name} 一致 (Max Diff: {max_diff:.6e})")
        return True
    except AssertionError:
        max_diff = np.abs(x1 - x2).max()
        print(f"❌ {name} 不一致! Max Diff: {max_diff:.6e}")
        return False


# ------------------------------------------------------------------
# 3. 导入模块
# ------------------------------------------------------------------
from rwkv_ops import rwkv7_op_inference
from rwkv_ops.rwkv7_kernel.native_keras_op import generalized_delta_rule

print("=" * 60)
print("无 Mask 推理测试")
print("=" * 60)

# ------------------------------------------------------------------
# 3.1 无 Mask 前向测试
# ------------------------------------------------------------------
cuda_out, cuda_state = rwkv7_op_inference(
    r=r, k=k, v=v, a=a, b=b, w=w, initial_state=h0
)

# Native 版本（detach 副本）
r_n = ops.copy(r)
k_n = ops.copy(k)
v_n = ops.copy(v)
a_n = ops.copy(a)
b_n = ops.copy(b)
w_n = ops.copy(w)
h0_n = ops.copy(h0)

native_out, native_state = generalized_delta_rule(
    r=r_n, k=k_n, v=v_n, a=a_n, b=b_n, w=w_n, initial_state=h0_n
)

test_is_close("fwd_pred", native_out, cuda_out, atol=1e-5, rtol=1e-2)
test_is_close("fwd_state", native_state, cuda_state, atol=1e-5, rtol=1e-3)
print("🎉 无 Mask 推理测试完成")

# ------------------------------------------------------------------
# 4. 带 Mask 推理测试
# ------------------------------------------------------------------
print("\n" + "=" * 60)
print("带 Mask 推理测试")
print("=" * 60)

# 构造随机 Mask（30% 冻结）
mask_np = np.ones((B, T), dtype=np.float32)
freeze_indices = np.random.rand(B, T) < 0.3
mask_np[freeze_indices] = 0.0
mask = jnp.asarray(mask_np)

print(f"Mask 冻结比例: {(mask == 0).mean():.2%}")

# 为了保证数值稳定性，重新生成一份输入（避免之前的影响）
inputs_mask = [np.random.randn(B, T, H, K) for _ in range(30)]
jax_inputs_mask = [jnp.asarray(t, "bfloat16") for t in inputs_mask]

a_m = -normalize(jax_inputs_mask[3], dim=-1, p=2.0)
b_m = normalize(jax_inputs_mask[3], dim=-1, p=2.0)
w_m = jax_inputs_mask[4]
r_m = jax_inputs_mask[0]
k_m = jax_inputs_mask[1]
v_m = jax_inputs_mask[2]
w_m = -ops.softplus(w_m) - 0.5
h0_m = jnp.asarray(np.random.randn(B, H, K, K), "float32")

# Native 副本
r_m_n = ops.copy(r_m)
k_m_n = ops.copy(k_m)
v_m_n = ops.copy(v_m)
a_m_n = ops.copy(a_m)
b_m_n = ops.copy(b_m)
w_m_n = ops.copy(w_m)
h0_m_n = ops.copy(h0_m)

# ------------------------------------------------------------------
# 4.1 带 Mask 前向测试
# ------------------------------------------------------------------
cuda_out_mask, cuda_state_mask = rwkv7_op_inference(
    r=r_m, k=k_m, v=v_m, a=a_m, b=b_m, w=w_m, initial_state=h0_m, mask=mask
)

native_out_mask, native_state_mask = generalized_delta_rule(
    r=r_m_n,
    k=k_m_n,
    v=v_m_n,
    a=a_m_n,
    b=b_m_n,
    w=w_m_n,
    initial_state=h0_m_n,
    mask=mask,
)

print("\n>>> 随机 Mask 前向对比")
test_is_close("fwd_pred_mask", native_out_mask, cuda_out_mask, atol=1e-5, rtol=1e-2)
test_is_close(
    "fwd_state_mask", native_state_mask, cuda_state_mask, atol=1e-5, rtol=1e-3
)

# ------------------------------------------------------------------
# 5. Mask 物理一致性验证
# ------------------------------------------------------------------
print("\n" + "=" * 60)
print("Mask 物理一致性验证")
print("=" * 60)

# 测试 1: 全 0 Mask，状态应该完全不变
print("\n[测试 1] 全 0 Mask（状态冻结）")
mask_all_zero = jnp.zeros((B, T), jnp.float32)
h0_frozen = h0_m.copy()

out_frozen, state_frozen = rwkv7_op_inference(
    r=r_m,
    k=k_m,
    v=v_m,
    a=a_m,
    b=b_m,
    w=w_m,
    initial_state=h0_frozen,
    mask=mask_all_zero,
)

diff = jnp.abs(state_frozen - h0_frozen).max()
if diff < 1e-5:
    print(f"✅ 状态冻结验证通过 (Max Diff: {diff:.2e})")
else:
    print(f"❌ 状态被错误改变 (Max Diff: {diff:.2e})")

# 测试 2: 全 1 Mask，应与无 Mask 版本一致（使用相同输入）
print("\n[测试 2] 全 1 Mask（与无 Mask 等价）")
mask_all_one = jnp.ones((B, T), jnp.float32)

# 使用无 mask 版本的输入重新测试
out_one_mask, state_one_mask = rwkv7_op_inference(
    r=r, k=k, v=v, a=a, b=b, w=w, initial_state=h0, mask=mask_all_one
)

# 注意：这里与之前的 cuda_out/cuda_state 比较（无 mask 版本）
pred_diff = jnp.abs(out_one_mask - cuda_out).max()
state_diff = jnp.abs(state_one_mask - cuda_state).max()

if pred_diff < 1e-5 and state_diff < 1e-5:
    print("✅ 全 1 Mask 与无 Mask 等价")

else:
    print(f"⚠️ 存在微小差异 (输出: {pred_diff:.2e}, 状态: {state_diff:.2e})")

# 测试 3: 边界情况 - 只在最后一个时间步更新
print("\n[测试 3] 仅最后一帧更新（前 T-1 帧冻结）")
mask_last = np.zeros((B, T), dtype=np.float32)
mask_last[:, -1] = 1.0  # 只有最后一步更新
mask_last = jnp.asarray(mask_last)

out_last, state_last = rwkv7_op_inference(
    r=r_m, k=k_m, v=v_m, a=a_m, b=b_m, w=w_m, mask=mask_last, initial_state=h0_m
)

# 验证：最终状态应该与初始状态有差异（因为最后一步更新了）
state_change = jnp.abs(state_last - h0_m).mean()
if state_change > 0:
    print(f"✅ 最后一帧更新验证通过 (状态平均变化: {state_change:.4f})")
else:
    print("⚠️ 状态未变化，可能存在问题")

print("\n" + "=" * 60)
print("🎉🎉🎉 全部 JAX 推理测试完成 🎉🎉🎉")
print("=" * 60)



"""
JAX 版 RWKV7 单步 wkv kernel（仅前向传播）
延迟编译 CUDA 扩展，专为 T=1 场景优化
"""

from __future__ import annotations
import pathlib
import subprocess
import ctypes
import jax
import jax.numpy as jnp
from typing import Optional, Tuple, Union

# ---------- 延迟编译（改到当前目录） ----------
_CURRENT_DIR = pathlib.Path(__file__).parent.absolute()


def get_jax_generalized_delta_rule_single_step(HEAD_SIZE=64):
    _BUILD_DIR = _CURRENT_DIR / f"build_single_step_{HEAD_SIZE}"
    _SO_PATH = _CURRENT_DIR / f"build_single_step_{HEAD_SIZE}/wkv7_single_step.so"

    def _ensure_compiled() -> pathlib.Path:
        """首次调用时编译 CUDA 扩展，产出放在当前源码目录"""
        if _SO_PATH.exists():
            return _SO_PATH

        print("[rwkv7_single_step_jax] First use – compiling CUDA kernel…")
        src_dir = _CURRENT_DIR
        build_dir = _BUILD_DIR
        build_dir.mkdir(exist_ok=True)

        # 获取 XLA 头文件路径
        xla_include_dir = jax.ffi.include_dir()
        if not xla_include_dir:
            raise RuntimeError("jax.ffi.include_dir() 返回空，请检查 JAX >= 0.4.31")

        # CUDA 编译 flags（移除 CHUNK_LEN 定义）
        cuda_flags = [
            "-ftz=true",
            "-prec-div=false",
            "-prec-sqrt=false",
            "--use_fast_math",
            "-O3",
            "-Xptxas=-O3",
            "-res-usage",
            "--extra-device-vectorization",
            f"-D_C_={HEAD_SIZE}",
        ]

        # CMake 配置
        cmake_args = [
            "cmake",
            "-S",
            str(src_dir),
            "-B",
            str(build_dir),
            "-DCMAKE_BUILD_TYPE=Release",
            f"-DCMAKE_INSTALL_PREFIX={_CURRENT_DIR}",
            f"-DXLA_INCLUDE_DIR={xla_include_dir}",
            f"-DCMAKE_CUDA_FLAGS={' '.join(cuda_flags)}",
        ]
        subprocess.check_call(cmake_args)

        # 构建与安装
        subprocess.check_call(["cmake", "--build", str(build_dir), "-j"])
        subprocess.check_call(["cmake", "--install", str(build_dir)])

        if not _SO_PATH.exists():
            raise RuntimeError("Compilation failed – wkv7_single_step.so not found.")

        print("[rwkv7_single_step_jax] Compilation finished – output at", _SO_PATH)
        return _SO_PATH

    # 注册 FFI 符号（仅前向）
    _lib = ctypes.CDLL(_ensure_compiled())
    jax.ffi.register_ffi_target(
        "wkv7_single_step_fwd",
        jax.ffi.pycapsule(_lib.Wkv7SingleStepFwd),
        platform="CUDA",
    )

    # ---------- 工具 ----------
    def _transpose_head(x: jnp.ndarray, head_first: bool) -> jnp.ndarray:
        """(B, 1, H, K) <-> (B, H, 1, K)"""
        if head_first:
            return jnp.transpose(x, (0, 2, 1, 3))
        x = jnp.asarray(x, dtype=jnp.bfloat16)
        return x

    # ---------- 前向 kernel ----------
    def _wkv7_single_step_kernel(
        w: jnp.ndarray,
        q: jnp.ndarray,
        k: jnp.ndarray,
        v: jnp.ndarray,
        a: jnp.ndarray,
        b: jnp.ndarray,
        h0: jnp.ndarray,
    ):
        """
        内部 kernel 接口
        参数: w,q,k,v,a,b,h0 -> y,s
        """
        B, H, K = q.shape
        dtype = q.dtype

        out_type = jax.ShapeDtypeStruct((B, H, K), dtype)
        s_type = jax.ShapeDtypeStruct((B, H, K, K), jnp.float32)

        y, s = jax.ffi.ffi_call(
            "wkv7_single_step_fwd", (out_type, s_type), vmap_method="broadcast_all"
        )(w, q, k, v, a, b, h0)

        return y, s

    def wk7_single_step_kernel(
        w: jnp.ndarray,
        q: jnp.ndarray,
        k: jnp.ndarray,
        v: jnp.ndarray,
        a: jnp.ndarray,
        b: jnp.ndarray,
        h0: jnp.ndarray,
    ):
        """前向计算函数"""
        y, s = _wkv7_single_step_kernel(w, q, k, v, a, b, h0)
        final_state = s  # 单步后直接返回状态
        return (y, final_state)

    # ---------- 主接口 ----------
    def generalized_delta_rule_single_step(
        r: jnp.ndarray,
        w: jnp.ndarray,
        k: jnp.ndarray,
        v: jnp.ndarray,
        a: jnp.ndarray,
        b: jnp.ndarray,
        initial_state: Optional[jnp.ndarray] = None,
        output_final_state: bool = True,
        head_first: bool = False,
    ) -> Union[jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray]]:
        """
        单步广义 delta 规则（仅前向）
        参数:
            r,w,k,v,a,b: 输入张量，形状必须为 (B, 1, H, K) 或 (B, H, 1, K)
            initial_state: 可选 (B, H, K, K) 初始状态，None 则零初始化
            output_final_state: 是否同时返回最后状态
            head_first: 是否将 head 维提前
        返回:
            out: (B, 1, H, K)  与输入 dtype 一致
            last_state: (B, H, K, K) 当 output_final_state=True
        """
        # 统一转 (B, 1, H, K) 并验证 T=1
        r = _transpose_head(r, head_first)
        w = _transpose_head(w, head_first)
        k = _transpose_head(k, head_first)
        v = _transpose_head(v, head_first)
        a = _transpose_head(a, head_first)
        b = _transpose_head(b, head_first)

        B, T, H, K = r.shape
        if T != 1:
            raise ValueError(f"Single-step kernel requires T=1, but got T={T}.")

        # 处理初始状态
        if initial_state is None:
            h0 = jnp.zeros((B, H, K, K), jnp.float32)
        else:
            h0 = jnp.asarray(initial_state, jnp.float32)

        # 移除 T 维度后调用 kernel
        r = r[:, 0, :, :]  # (B, H, K)
        w = w[:, 0, :, :]
        k = k[:, 0, :, :]
        v = v[:, 0, :, :]
        a = a[:, 0, :, :]
        b = b[:, 0, :, :]

        # 调用前向 kernel
        out, last_state = wk7_single_step_kernel(w, r, k, v, a, b, h0)

        # 恢复 T 维度
        out = jnp.expand_dims(out, axis=1)  # (B, 1, H, K)
        out = jnp.asarray(out, r.dtype)

        if output_final_state:
            return out, last_state
        return out

    return generalized_delta_rule_single_step



import os
import argparse

parser = argparse.ArgumentParser(
    description="Test RWKV7 Op with different backends and kernel types."
)
parser.add_argument(
    "--backend",
    type=str,
    choices=["torch", "jax", "numpy", "tensorflow", "openvino"],
    required=True,
)
args = parser.parse_args()

os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
os.environ["KERAS_BACKEND"] = args.backend
os.environ["KERNEL_TYPE"] = "cuda"

import numpy as np
from keras import ops

# ------------------------------------------------------------------
# 1. 构造输入
# ------------------------------------------------------------------
T = 512
B = 5
H = 6
K = 64
inputs = [np.random.randn(B, T, H, K) for _ in range(30)]
tf_inputs = [ops.convert_to_tensor(t, "bfloat16") for t in inputs]


def normalize(
    z,
    p=2,
    dim=-1,
    eps: float = 1e-12,
):
    # F.normalize like api
    denom = ops.norm(z, ord=p, axis=dim, keepdims=True)
    denom = ops.maximum(denom, 1e-12)
    return z / denom


mask = ops.concatenate(
    [ops.zeros((B, T // 2, 1, 1)), ops.ones((B, T // 2, 1, 1))], axis=1
)
mask = ops.cast(mask, "bfloat16")

a = -normalize(tf_inputs[3], dim=-1, p=2.0) * mask
b = normalize(tf_inputs[3], dim=-1, p=2.0) * mask

w = tf_inputs[4] * mask  # decay / gate
r = tf_inputs[0] * mask  # receptance
k = tf_inputs[1] * mask
v = tf_inputs[2] * mask
w = -ops.softplus(w) - 0.5
w = ops.where(mask, w, -1e9)
h0 = ops.convert_to_tensor(np.random.randn(B, H, K, K), "float32")
# ------------------------------------------------------------------
# 2. CUDA 版本前向 + 反向
# ------------------------------------------------------------------
from rwkv_ops import rwkv7_op


cuda_out, cuda_state = rwkv7_op(r=r, k=k, v=v, a=a, b=b, w=w, initial_state=h0)

r_n = ops.copy(r)
k_n = ops.copy(k)
v_n = ops.copy(v)
a_n = ops.copy(a)
b_n = ops.copy(b)
w_n = ops.copy(w)
h0_n = ops.copy(h0)


def transpose_head(x, head_first):
    """
    对输入张量进行转置操作。

    参数:
    x: 输入张量。
    head_first: 布尔值，决定是否进行转置。

    返回:
    转置后的张量（如果head_first为True），否则返回原张量。
    """
    x = ops.cast(x, "float32")
    if head_first:
        return ops.transpose(x, (0, 2, 1, 3))
    else:
        return x


def generalized_delta_rule(
    r,
    w,
    k,
    v,
    a,
    b,
    initial_state=None,
    output_final_state: bool = True,
    head_first: bool = False,
):
    """
    实现广义delta规则的函数。

    参数:
    r: 输入张量。
    w: 权重张量。
    k, v, a, b: 其他输入张量。
    initial_state: 初始状态张量。
    output_final_state: 是否输出最终状态。
    head_first: 是否在计算中将head维度放在第一位。

    返回:
    根据output_final_state参数决定是否返回最终状态。
    """
    DTYPE = r.dtype
    B, T, H, N = ops.shape(r)
    r = transpose_head(r, head_first)

    k = transpose_head(k, head_first)

    v = transpose_head(v, head_first)
    a = transpose_head(a, head_first)
    b = transpose_head(b, head_first)
    w = transpose_head(w, head_first)
    w = ops.exp(-ops.exp(w))

    if initial_state is not None:
        state = initial_state
        if ops.shape(state)[0] == 1:
            state = ops.broadcast_to(state, (B, H, N, N))
    else:
        state = ops.zeros((B, H, N, N))
    state = ops.cast(state, "float32")
    out = ops.zeros((B, T, H, N), DTYPE)

    def step(t, inputs):
        """
        执行单个时间步的计算。

        参数:
        t: 当前时间步。
        inputs: 包含当前状态和输出的列表。

        返回:
        更新后的状态和输出。
        """
        state, out = inputs
        kk = ops.reshape(k[:, t, :], (B, H, 1, N))
        rr = ops.reshape(r[:, t, :], (B, H, N, 1))
        vv = ops.reshape(v[:, t, :], (B, H, N, 1))
        aa = ops.reshape(a[:, t, :], (B, H, N, 1))
        bb = ops.reshape(b[:, t, :], (B, H, 1, N))
        state = state * w[:, t, :, None, :] + state @ aa @ bb + vv @ kk
        o = ops.cast((state @ rr), out.dtype)
        out = ops.slice_update(out, [0, t, 0, 0], ops.reshape(o, (B, 1, H, N)))
        return [state, out]

    state, out = ops.fori_loop(0, T, step, [state, out])

    if output_final_state:
        return ops.cast(out, DTYPE), state
    return ops.cast(out, DTYPE)


native_out, native_state = generalized_delta_rule(
    r=r_n, k=k_n, v=v_n, a=a_n, b=b_n, w=w_n, initial_state=h0_n
)


def test_is_close(name, x1, x2, atol=2.5e-2, rtol=1e-3):
    x1 = ops.convert_to_numpy(ops.cast(x1, "float32"))
    x2 = ops.convert_to_numpy(ops.cast(x2, "float32"))
    if np.sum(np.isnan(x1)) == 0 and np.sum(np.isnan(x2)) == 0:
        print(f"✅✅{name} 不存在nan✅✅")
    else:
        print(f"❌❌{name} 你妈的有nan❌❌")
    if np.sum(np.abs(x1 - x2)) < 1e-4:
        print(f"✅✅{name} 输出结果完全一致✅✅")
        return
    try:
        np.testing.assert_allclose(
            x1,
            x2,
            atol=atol,
            rtol=rtol,
        )
        print(f"✅ {name} 梯度一致")
    except AssertionError as e:
        print(f"❌ {name} 梯度不一致")
        print(e)


test_is_close("fwd_pred", native_out, cuda_out, atol=1e-5, rtol=1e-2)
test_is_close("fwd_state", native_state, cuda_state, atol=1e-5, rtol=1e-3)
print(f"🎉🎉🎉🎉test_script/test_rwkv7_native.py在{args.backend}后端的测试结束🎉🎉🎉🎉")



from ..triton_kernel.mhc_post_op import *
import jax
import jax.numpy as jnp
import jax_triton as jt
import jax.tree_util as jtu

# 引入自定义分区器和切分规则 (适配 JAX 新版 Shardy 引擎)
from jax.experimental.custom_partitioning import custom_partitioning

# =========================================================================
# 【核心配置】：为 Shardy 引擎定义的静态 Einsum 切分映射字符串
# 字母含义: b=Batch, t=Time, c=Channel, n=NSize1, m=NSize2
# =========================================================================
BATCH_AXIS_NAME = "data"

# FWD_RULE 含义:
# 输入1(layer_out):  b t c
# 输入2(x_expanded): b t n c
# 输入3(h_post_raw): b t n
# 输入4(H_res):      b t n m (用独立的 m 避免重复字母报错)
# 输出1(out):        b t n c
FWD_RULE = "b t c, b t n c, b t n, b t n m -> b t n c"

# BWD_RULE 含义:
# 输入1-4同上, 输入5(grad_out): b t n c
# 输出1(gx):       b t c
# 输出2(gh_res):   b t n c
# 输出3(gh_pre):   b t n
# 输出4(grad_H):   b t n m
BWD_RULE = "b t c, b t n c, b t n, b t n m, b t n c -> b t c, b t n c, b t n, b t n m"


# 兼容老版本的推导函数
def _fwd_infer_sharding(arg_shapes, arg_shardings):
    qs = arg_shardings[1]  # 输出与 x_expanded (索引1) shape/sharding 相同
    return (qs,)


def _bwd_infer_sharding(arg_shapes, arg_shardings):
    # 输出梯度对应前 4 个输入
    return (arg_shardings[0], arg_shardings[1], arg_shardings[2], arg_shardings[3])


# JAX 官方标准的 partition 回调生成器
def _create_partition(impl_fn):
    def partition(mesh, arg_shapes, result_shape):
        def lower_fn(*args):
            return impl_fn(*args)

        result_shardings = jtu.tree_map(lambda x: x.sharding, result_shape)
        arg_shardings = jtu.tree_map(lambda x: x.sharding, arg_shapes)
        return mesh, lower_fn, result_shardings, arg_shardings

    return partition


# =========================================================================


# --- 1. 前向 Launcher ---
def mhc_post_op_fwd_kernel_call(
    layer_out: jax.Array, x_expanded: jax.Array, h_post_raw: jax.Array, H_res: jax.Array
):
    batch, time, NSIZE, channel = x_expanded.shape
    total_bt = batch * time

    # 1. 重塑形状
    x_v = x_expanded.reshape(total_bt, NSIZE, channel)
    h_v = h_post_raw.reshape(total_bt, NSIZE)
    H_v = H_res.reshape(total_bt, NSIZE, NSIZE)
    l_v = layer_out.reshape(total_bt, channel)

    # 2. 计算步长
    sx_bt, sx_n, sx_c = jt.strides_from_shape(x_v.shape)
    sh_bt, sh_n = jt.strides_from_shape(h_v.shape)
    sH_bt, sH_n1, sH_n2 = jt.strides_from_shape(H_v.shape)
    sl_bt, sl_c = jt.strides_from_shape(l_v.shape)

    # 输出形状与 x_v 一致
    out_struct = jax.ShapeDtypeStruct(x_v.shape, x_v.dtype)
    so_bt, so_n, so_c = sx_bt, sx_n, sx_c

    # 3. 定义 Grid
    grid = lambda meta: (total_bt, jt.cdiv(channel, meta["BLOCK_CHANNEL"]))

    # 4. 调用 Triton
    out_v = jt.triton_call(
        x_v,
        h_v,
        H_v,
        l_v,  # 输入 Array
        kernel=mhc_fused_forward_kernel,
        out_shape=out_struct,
        grid=grid,
        # 标量/步长参数
        stride_output_batch_time=so_bt,
        stride_output_n_size=so_n,
        stride_output_channel=so_c,
        stride_x_batch_time=sx_bt,
        stride_x_n_size=sx_n,
        stride_x_channel=sx_c,
        stride_h_batch_time=sh_bt,
        stride_h_n_size=sh_n,
        stride_H_batch_time=sH_bt,
        stride_H_n_size_1=sH_n1,
        stride_H_n_size_2=sH_n2,
        stride_layer_out_batch_time=sl_bt,
        stride_layer_out_channel=sl_c,
        CHANNEL_SIZE=channel,
        NSIZE=NSIZE,
    )

    return out_v.reshape(batch, time, NSIZE, channel)


# =========================================================================
# 【新增】：包装前向算子 (适配 SPMD 并行)
# =========================================================================
@custom_partitioning
def mhc_post_op_fwd_spmd(layer_out, x_expanded, h_post_raw, H_res):
    return mhc_post_op_fwd_kernel_call(layer_out, x_expanded, h_post_raw, H_res)


mhc_post_op_fwd_spmd.def_partition(
    infer_sharding_from_operands=_fwd_infer_sharding,
    sharding_rule=FWD_RULE,
    partition=_create_partition(mhc_post_op_fwd_kernel_call),
)


# --- 2. 反向 Launcher ---
# 【修改】：展平输入参数为 5 个独立 Tensor，方便 SPMD 推导分片规则
def mhc_post_op_bwd_kernel_call(layer_out, x_expanded, h_post_raw, H_res, grad_output):
    B, T, n, C = x_expanded.shape
    total_bt = B * T

    # 1. 准备输入
    x_v = x_expanded.reshape(total_bt, n, C)
    h_v = h_post_raw.reshape(total_bt, n)
    H_v = H_res.reshape(total_bt, n, n)
    l_v = layer_out.reshape(total_bt, C)
    g_out_v = grad_output.reshape(total_bt, n, C)

    # 2. 计算输入步长
    sx_bt, sx_n, sx_c = jt.strides_from_shape(x_v.shape)
    sh_bt, sh_n = jt.strides_from_shape(h_v.shape)
    sH_bt, sH_n1, sH_n2 = jt.strides_from_shape(H_v.shape)
    sl_bt, sl_c = jt.strides_from_shape(l_v.shape)
    sg_bt, sg_n, sg_c = jt.strides_from_shape(g_out_v.shape)

    # 3. 准备输出结构
    out_shapes = [
        jax.ShapeDtypeStruct(x_v.shape, x_v.dtype),  # grad_x
        jax.ShapeDtypeStruct(h_v.shape, jnp.float32),  # grad_h
        jax.ShapeDtypeStruct(H_v.shape, jnp.float32),  # grad_H
        jax.ShapeDtypeStruct(l_v.shape, l_v.dtype),  # grad_l
    ]

    # 4. 计算输出步长
    sgx_bt, sgx_n, sgx_c = sx_bt, sx_n, sx_c
    sgh_bt, sgh_n = sh_bt, sh_n
    sgH_bt, sgH_n1, sgH_n2 = sH_bt, sH_n1, sH_n2
    sgl_bt, sgl_c = sl_bt, sl_c

    # 5. 调用 Triton
    grad_x_v, grad_h_v, grad_H_v, grad_l_v = jt.triton_call(
        x_v,
        h_v,
        H_v,
        l_v,
        g_out_v,
        kernel=mhc_fused_backward_kernel,
        out_shape=out_shapes,
        grid=(total_bt, 1),
        stride_x_bt=sx_bt,
        stride_x_n=sx_n,
        stride_x_c=sx_c,
        stride_h_bt=sh_bt,
        stride_h_n=sh_n,
        stride_H_bt=sH_bt,
        stride_H_n1=sH_n1,
        stride_H_n2=sH_n2,
        stride_l_bt=sl_bt,
        stride_l_c=sl_c,
        stride_g_bt=sg_bt,
        stride_g_n=sg_n,
        stride_g_c=sg_c,
        stride_gx_bt=sgx_bt,
        stride_gx_n=sgx_n,
        stride_gx_c=sgx_c,
        stride_gl_bt=sgl_bt,
        stride_gl_c=sgl_c,
        stride_gh_bt=sgh_bt,
        stride_gh_n=sgh_n,
        stride_gH_bt=sgH_bt,
        stride_gH_n1=sgH_n1,
        stride_gH_n2=sgH_n2,
        CHANNEL_SIZE=C,
        NSIZE=n,
    )

    # 6. 返回梯度 (注意返回顺序需要和原函数输入参数一致：layer_out, x_expanded, h_post_raw, H_res)
    return (
        grad_l_v.reshape(B, T, C),
        grad_x_v.reshape(B, T, n, C),
        grad_h_v.reshape(B, T, n).astype(h_post_raw.dtype),
        grad_H_v.reshape(B, T, n, n).astype(H_res.dtype),
    )


# =========================================================================
# 【新增】：包装反向算子 (适配 SPMD 并行)
# =========================================================================
@custom_partitioning
def mhc_post_op_bwd_spmd(layer_out, x_expanded, h_post_raw, H_res, grad_output):
    return mhc_post_op_bwd_kernel_call(
        layer_out, x_expanded, h_post_raw, H_res, grad_output
    )


mhc_post_op_bwd_spmd.def_partition(
    infer_sharding_from_operands=_bwd_infer_sharding,
    sharding_rule=BWD_RULE,
    partition=_create_partition(mhc_post_op_bwd_kernel_call),
)


# --- JAX 接口绑定 ---


@jax.custom_vjp
def mhc_post_op(layer_out, x_expanded, h_post_raw, H_res):
    # 【修改】：调用切分规则封装后的 FWD
    return mhc_post_op_fwd_spmd(
        layer_out.astype(jnp.bfloat16),
        x_expanded.astype(jnp.bfloat16),
        h_post_raw.astype(jnp.float32),
        H_res.astype(jnp.float32),
    )


def mhc_post_op_fwd(layer_out, x_expanded, h_post_raw, H_res):
    out = mhc_post_op(layer_out, x_expanded, h_post_raw, H_res)
    return out, (layer_out, x_expanded, h_post_raw, H_res)


def mhc_post_op_bwd(res, grad_output):
    layer_out, x_expanded, h_post_raw, H_res = res

    # 保证健壮性，若上游传了 None 梯度则初始化为零张量
    if grad_output is None:
        grad_output = jnp.zeros_like(x_expanded)

    # 【修改】：调用切分规则封装后的 BWD，展平传递参数
    grads = mhc_post_op_bwd_spmd(layer_out, x_expanded, h_post_raw, H_res, grad_output)
    return grads


mhc_post_op.defvjp(mhc_post_op_fwd, mhc_post_op_bwd)



import torch
from .triton_kernel import *


class TritonWindBackstepping(torch.autograd.Function):
    @staticmethod
    def forward(ctx, w, q, k, v, a, b, h0):
        # 确保输入是标准的 [B, N_HEAD, T_LEN, H_SIZE] 且连续
        # w, q, k, v, a, b 已经在 generalized_delta_rule 中处理过连续性
        B, N, T, H = w.shape
        DTYPE = q.dtype
        CHUNK_LEN = 16

        if T % CHUNK_LEN != 0:
            raise ValueError(f"Sequence length T={T} must be divisible by {CHUNK_LEN}")
        if H != 64:
            raise ValueError(
                f"Triton kernel currently only supports Head Size = 64, got {H}"
            )

        # 1. 分配输出与中间变量
        # OUT 使用输入相同类型 (通常 bf16), SA 和 CHKP 强制使用 fp32 保证梯度精度
        out = torch.empty_like(v)
        sa_out = torch.empty(B, N, T, H, dtype=torch.float32, device=w.device)
        state_chkp = torch.empty(
            B, N, T // CHUNK_LEN, H, H, dtype=torch.float32, device=w.device
        )

        # 定义 Grid 分块策略
        def grid(META):
            return ((B + META["MINI_BSZ"] - 1) // META["MINI_BSZ"], N)

        # 2. 启动前向内核
        rwkv7_fwd_kernel[grid](
            R=q,
            W=w,
            K=k,
            V=v,
            A=a,
            B_param=b,
            H0=h0,
            OUT=out,
            SA_OUT=sa_out,
            STATE_CHKP=state_chkp,
            N_HEAD=N,
            B_BATCH=B,
            T_LEN=T,
            H_SIZE=H,
            CHUNK_LEN=CHUNK_LEN,
        )

        # 3. 保存反向传播所需的张量
        ctx.save_for_backward(w, q, k, v, a, b, state_chkp, sa_out)

        # 4. 获取并返回最终状态
        # state_chkp 最后一位就是最终时刻的状态
        last_state = state_chkp[:, :, -1, :, :].clone()

        return out.to(DTYPE), last_state

    @staticmethod
    def backward(ctx, dy, dht):
        # dy: [B, N, T, H] (对应输出 y 的梯度)
        # dht: [B, N, H, H] (对应最终状态的梯度)

        # 1. 恢复前向变量
        w, q, k, v, a, b, state_chkp, sa_out = ctx.saved_tensors
        B, N, T, H = w.shape
        DTYPE = q.dtype
        CHUNK_LEN = 16

        # 2. 准备梯度输出张量
        dw = torch.empty_like(w)
        dq = torch.empty_like(q)
        dk = torch.empty_like(k)
        dv = torch.empty_like(v)
        da = torch.empty_like(a)
        db = torch.empty_like(b)
        # dh0 始终使用 float32
        dh0 = torch.empty(B, N, H, H, dtype=torch.float32, device=w.device)

        # 3. 预处理输入梯度
        dy = dy.contiguous()
        if dht is None:
            dht = torch.zeros(B, N, H, H, dtype=torch.float32, device=w.device)
        else:
            dht = dht.contiguous().to(torch.float32)

        # 定义 Grid
        def grid(META):
            return ((B + META["MINI_BSZ"] - 1) // META["MINI_BSZ"], N)

        # 4. 启动反向内核
        rwkv7_bwd_kernel[grid](
            R=q,
            W=w,
            K=k,
            V=v,
            A=a,
            B_param=b,
            SA=sa_out,
            STATE_CHKP=state_chkp,
            DY=dy,
            DHT=dht,
            B_BATCH=B,
            N_HEAD=N,
            T_LEN=T,
            DR=dq,
            DW=dw,
            DK=dk,
            DV=dv,
            DA=da,
            DB=db,
            DH0=dh0,
            H_SIZE=H,
            CHUNK_LEN=CHUNK_LEN,
        )

        # 5. 返回各输入的梯度 (必须与 forward 的输入顺序一致)
        # 返回顺序: w, q, k, v, a, b, h0
        return (
            dw.to(DTYPE),
            dq.to(DTYPE),
            dk.to(DTYPE),
            dv.to(DTYPE),
            da.to(DTYPE),
            db.to(DTYPE),
            dh0,
        )


class TritonWindBacksteppingWithMask(torch.autograd.Function):
    @staticmethod
    def forward(ctx, w, q, k, v, a, b, mask, h0):
        # 确保输入是标准的连续张量 [B, N_HEAD, T_LEN, H_SIZE]
        B, N, T, H = w.shape
        DTYPE = q.dtype
        CHUNK_LEN = 16

        if T % CHUNK_LEN != 0:
            raise ValueError(f"Sequence length T={T} must be divisible by {CHUNK_LEN}")
        if H != 64:
            raise ValueError(
                f"Triton kernel currently only supports Head Size = 64, got {H}"
            )

        # mask 必须是连续的张量，通常传入的形状可能是 [B, T] 或 [B, T, 1, 1]
        # Triton kernel 中我们按照 [B, T] 连续内存读取，因此如果维度多余需要 squeeze，然后 contiguous
        if mask.dim() > 2:
            mask = mask.view(B, T)
        mask = mask.contiguous().to(torch.float32)  # Triton内核中统一转为 fp32 计算

        # 分配输出与中间变量
        out = torch.empty_like(v)
        sa_out = torch.empty(B, N, T, H, dtype=torch.float32, device=w.device)
        state_chkp = torch.empty(
            B, N, T // CHUNK_LEN, H, H, dtype=torch.float32, device=w.device
        )

        # 定义 Grid 分块策略
        def grid(META):
            return ((B + META["MINI_BSZ"] - 1) // META["MINI_BSZ"], N)

        # 启动带 Mask 的前向内核
        rwkv7_fwd_kernel_with_mask[grid](
            R=q,
            W=w,
            K=k,
            V=v,
            A=a,
            B_param=b,
            MASK=mask,
            H0=h0,
            B_BATCH=B,
            N_HEAD=N,
            T_LEN=T,
            OUT=out,
            SA_OUT=sa_out,
            STATE_CHKP=state_chkp,
            H_SIZE=H,
            CHUNK_LEN=CHUNK_LEN,
        )

        # 保存反向传播所需的张量 (注意要保存 mask)
        ctx.save_for_backward(w, q, k, v, a, b, mask, state_chkp, sa_out)

        # 获取并返回最终状态
        last_state = state_chkp[:, :, -1, :, :].clone()

        return out.to(DTYPE), last_state

    @staticmethod
    def backward(ctx, dy, dht):
        # 1. 恢复前向变量
        w, q, k, v, a, b, mask, state_chkp, sa_out = ctx.saved_tensors
        B, N, T, H = w.shape
        DTYPE = q.dtype
        CHUNK_LEN = 16

        # 2. 准备梯度输出张量
        dw = torch.empty_like(w)
        dq = torch.empty_like(q)
        dk = torch.empty_like(k)
        dv = torch.empty_like(v)
        da = torch.empty_like(a)
        db = torch.empty_like(b)
        dh0 = torch.empty(B, N, H, H, dtype=torch.float32, device=w.device)

        # 3. 预处理输入梯度
        dy = dy.contiguous()
        if dht is None:
            dht = torch.zeros(B, N, H, H, dtype=torch.float32, device=w.device)
        else:
            dht = dht.contiguous().to(torch.float32)

        # 定义 Grid
        def grid(META):
            return ((B + META["MINI_BSZ"] - 1) // META["MINI_BSZ"], N)

        # 4. 启动带 Mask 的反向内核
        rwkv7_bwd_kernel_with_mask[grid](
            R=q,
            W=w,
            K=k,
            V=v,
            A=a,
            B_param=b,
            MASK=mask,
            SA=sa_out,
            STATE_CHKP=state_chkp,
            B_BATCH=B,
            N_HEAD=N,
            T_LEN=T,
            DY=dy,
            DHT=dht,
            DR=dq,
            DW=dw,
            DK=dk,
            DV=dv,
            DA=da,
            DB=db,
            DH0=dh0,
            H_SIZE=H,
            CHUNK_LEN=CHUNK_LEN,
        )

        # 5. 返回梯度
        # 注意：forward 的输入签名是 (ctx, w, q, k, v, a, b, mask, h0)
        # 所以我们需要返回 8 个梯度：w, q, k, v, a, b, mask, h0
        # mask 不需要梯度，因此返回 None
        return (
            dw.to(DTYPE),
            dq.to(DTYPE),
            dk.to(DTYPE),
            dv.to(DTYPE),
            da.to(DTYPE),
            db.to(DTYPE),
            None,  # mask 的梯度为 None
            dh0,
        )


def generalized_delta_rule(
    r,
    w,
    k,
    v,
    a,
    b,
    initial_state=None,
    output_final_state: bool = True,
    head_first: bool = False,
    mask=None,
):
    """
    统一接口函数。
    支持输入形状:
        head_first=False: [B, T, N, H] (Keras 默认)
        head_first=True : [B, N, T, H] (Triton 算子标准)
    """
    if w.device.type != "cuda":
        raise RuntimeError("Triton kernel only supports CUDA devices.")

    # 1. 统一转换为 [B, N, T, H] 的 Head-First 连续张量
    if not head_first:
        r = r.transpose(1, 2)
        w = w.transpose(1, 2)
        k = k.transpose(1, 2)
        v = v.transpose(1, 2)
        a = a.transpose(1, 2)
        b = b.transpose(1, 2)
        # mask 处理 (通常为 [B, T] 或 [B, T, 1, 1])
        # 这里预留 Mask 转置逻辑，如果 mask 是 [B, T]，则不需要 transpose(1, 2)

    # 确保 Contiguous 极度重要，Triton 内核依赖于此进行偏移计算
    r, w, k, v, a, b = [x.contiguous() for x in [r, w, k, v, a, b]]

    B, N, T, H = w.shape

    # 2. 初始状态处理
    if initial_state is None:
        initial_state = torch.zeros(B, N, H, H, dtype=torch.float32, device=r.device)
    else:
        # 确保初始状态也是连续的 float32
        initial_state = initial_state.to(torch.float32).contiguous()

    # 3. 分发到 Autograd Function
    if mask is None:
        # 无 Mask 版本
        out, state = TritonWindBackstepping.apply(w, r, k, v, a, b, initial_state)
    else:
        if not mask.is_cuda:
            mask = mask.to(r.device)
        out, state = TritonWindBacksteppingWithMask.apply(
            w, r, k, v, a, b, mask, initial_state
        )

    # 4. 如果原始输入不是 head_first，则将输出转置回去 [B, N, T, H] -> [B, T, N, H]
    if not head_first:
        out = out.transpose(1, 2)

    return (out, state) if output_final_state else out



import os
import torch
import numpy as np

# 设置环境变量
os.environ["CUDA_VISIBLE_DEVICES"] = "1"

# 1. 导入算子
from rwkv_ops.mhc_kernel.native_op import (
    sinkhorn_knopp as native_sinkhorn,
    stream_aggregate as native_aggregate,
)
from rwkv_ops.mhc_kernel.torch_triton_op.mhc_pre_op import mhc_pre_op_fused

# ------------------------------------------------------------------
# 1. 配置参数与构造输入
# ------------------------------------------------------------------
B, T, n, C = 256, 256, 4, 768
device = "cuda"
dtype = torch.bfloat16


# 构造输入函数 (用于刷新梯度空间)
def get_inputs():
    x = torch.randn(B, T, n, C, device=device, dtype=dtype).requires_grad_(True)
    h_res = torch.randn(B, T, n, n, device=device, dtype=torch.float32).requires_grad_(
        True
    )
    h_pre = torch.randn(B, T, n, device=device, dtype=torch.float32).requires_grad_(
        True
    )
    return x, h_res, h_pre


# ------------------------------------------------------------------
# 2. 工具函数：数值比较
# ------------------------------------------------------------------
def test_is_close(name, x1, x2, atol=1e-2, rtol=1e-2):
    if x1 is None or x2 is None:
        print(f"❌ {name} 梯度不存在")
        return

    x1_np = x1.detach().cpu().float().numpy()
    x2_np = x2.detach().cpu().float().numpy()

    error = np.abs(x1_np - x2_np)
    avg_error = error.mean()
    rel_error = avg_error / (np.abs(x1_np).mean() + 1e-9)

    print(f"[{name}]")
    print(f"  平均绝对误差: {avg_error:.6e}")
    print(f"  平均相对误差: {rel_error:.6e}")

    equal_rate = np.sum(error < 1e-7) / x1_np.size * 100
    print(f"  {equal_rate:.2f}% 的数据完全一样")

    if np.isnan(x1_np).any() or np.isnan(x2_np).any():
        print("  ❌❌ 存在 NaN ❌❌")
        return

    try:
        np.testing.assert_allclose(x1_np, x2_np, atol=atol, rtol=rtol)
        print(f"  ✅ 数值一致 (max_diff: {error.max():.6e})")
    except AssertionError:
        print("  ❌ 数值不一致")
        # print(e)


# ------------------------------------------------------------------
# 3. 正确性校验 (Forward & Backward)
# ------------------------------------------------------------------
print("--- 开始数值正确性校验 ---")

# --- Native Pass ---
x_n, hr_n, hp_n = get_inputs()
h_res_n_out = native_sinkhorn(hr_n, num_iters=20)
x_in_n_out = native_aggregate(x_n, hp_n)

# 构造 Loss 以测试反向
loss_n = (x_in_n_out.float() ** 2).mean() + (h_res_n_out**2).mean()
loss_n.backward()

# --- Triton Pass ---
x_t, hr_t, hp_t = get_inputs()
# 拷贝同样的数值以确保起点一致
with torch.no_grad():
    x_t.copy_(x_n)
    hr_t.copy_(hr_n)
    hp_t.copy_(hp_n)

x_in_t_out, h_res_t_out = mhc_pre_op_fused(x_t, hr_t, hp_t, num_iters=20)

loss_t = (x_in_t_out.float() ** 2).mean() + (h_res_t_out**2).mean()
loss_t.backward()

# 比较前向
test_is_close("Forward: x_layer_in", x_in_n_out, x_in_t_out)
test_is_close("Forward: H_res", h_res_n_out, h_res_t_out)

# 比较反向梯度
test_is_close("Gradient: x", x_n.grad, x_t.grad)
test_is_close("Gradient: h_res_in", hr_n.grad, hr_t.grad)
test_is_close("Gradient: h_pre_in", hp_n.grad, hp_t.grad)

# ------------------------------------------------------------------
# 4. 性能基准测试 (Benchmark)
# ------------------------------------------------------------------
print("\n" + "=" * 40)
print("🚀 开始全流程性能基准测试 (Forward + Backward)")
print("=" * 40)

n_warmup = 10
n_repeat = 100


# Native 逻辑包装
def native_full_op(x, hr, hp):
    h_out = native_sinkhorn(hr, num_iters=20)
    x_out = native_aggregate(x, hp)
    return x_out, h_out


# torch.compile 逻辑包装
@torch.compile
def compiled_full_op(x, hr, hp):
    h_out = native_sinkhorn(hr, num_iters=20)
    x_out = native_aggregate(x, hp)
    return x_out, h_out


def benchmark_all(op_func, name, x, hr, hp):
    # 预热
    for _ in range(n_warmup):
        out_x, out_h = op_func(x, hr, hp)
        loss = (out_x.float() ** 2).mean() + (out_h**2).mean()
        loss.backward()
        x.grad, hr.grad, hp.grad = None, None, None
    torch.cuda.synchronize()

    start_event = torch.cuda.Event(enable_timing=True)
    end_event = torch.cuda.Event(enable_timing=True)

    start_event.record()
    for _ in range(n_repeat):
        out_x, out_h = op_func(x, hr, hp)
        loss = (out_x.float() ** 2).mean() + (out_h**2).mean()
        loss.backward()
        x.grad, hr.grad, hp.grad = None, None, None
    end_event.record()

    torch.cuda.synchronize()
    elapsed_time_ms = start_event.elapsed_time(end_event) / n_repeat
    print(f"👉 {name:20s} Full (Fwd+Bwd): {elapsed_time_ms:.4f} ms")
    return elapsed_time_ms


# 运行 Benchmark
x, hr, hp = get_inputs()
print(f"配置: B={B}, T={T}, n={n}, C={C}, dtype={dtype}")

t_native = benchmark_all(native_full_op, "Native (Python)", x, hr, hp)
t_compile = benchmark_all(compiled_full_op, "torch.compile", x, hr, hp)
t_triton = benchmark_all(mhc_pre_op_fused, "Triton Fused", x, hr, hp)

print("\n📊 性能提升总结:")
print(f"Triton vs Native:  {t_native / t_triton:.2f}x faster")
print(f"Triton vs Compile: {t_compile / t_triton:.2f}x faster")

print("\n🎉🎉 MHC Pre-Op 全校验结束 🎉🎉")
